# -*- coding: utf-8 -*-
import numpy as np
import pyqtgraph as pg
from qtpy import QtWidgets, QtCore, QtGui
from .. import global_vars as g
from .BaseProcess import BaseProcess, SliderLabel, WindowSelector,  MissingWindowError, CheckBox, ComboBox

__all__ = ['time_stamp','background','scale_bar']
     


class Time_Stamp(BaseProcess):
    """time_stamp(framerate,show=True)
    Adds a time stamp to a movie
    
    Parameters:
        | framerate (float) -- The number of frames per second
        | show (bool) -- Turns on or off the time stamp
    Returns:
        None
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        framerate=QtWidgets.QDoubleSpinBox()
        if hasattr(g.currentWindow,'framerate'):
            framerate.setValue(g.currentWindow.framerate)
        elif 'framerate' in g.settings.d.keys():
            framerate.setValue(g.settings['framerate'])
        framerate.setRange(0,1000000)
        framerate.setDecimals(10)
        show = CheckBox(); show.setChecked(True)
        self.items.append({'name':'framerate','string':'Frame Rate (Hz)','object':framerate})
        self.items.append({'name':'show','string':'Show','object':show})
        super().gui()
    def __call__(self,framerate,show=True,keepSourceWindow=None):
        w=g.currentWindow
        if show:
            w.framerate=framerate
            g.settings['framerate']=framerate
            if hasattr(w,'timeStampLabel') and w.timeStampLabel is not None:
                return
            w.timeStampLabel= pg.TextItem(html="<span style='font-size: 12pt;color:white;background-color:None;'>0 ms</span>")
            w.imageview.view.addItem(w.timeStampLabel)
            w.sigTimeChanged.connect(w.updateTimeStampLabel)
        else:
            if hasattr(w,'timeStampLabel') and w.timeStampLabel is not None:
                w.imageview.view.removeItem(w.timeStampLabel)
                w.timeStampLabel=None
                w.sigTimeChanged.disconnect(w.updateTimeStampLabel)
        return None
    def preview(self):
        framerate=self.getValue('framerate')
        show=self.getValue('show')
        self.__call__(framerate,show)

     
time_stamp=Time_Stamp()


class ShowCheckbox(CheckBox):

    def __init__(self, opacity_slider, parent=None):
        super().__init__(parent)
        self.stateChanged.connect(self.changed)
        self.opacity_slider = opacity_slider

    def changed(self, state):
        if state == 0:  # unchecked
            self.opacity_slider.setEnabled(True)
        if state == 2:  # checked
            self.opacity_slider.setEnabled(False)


class Background(BaseProcess):
    """ background(background_window, data_window)
    Overlays the background_window onto the data_window
    
    Parameters:
        | background_window (Window)
        | data_window (Window)
    Returns:
        None
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        background_window=WindowSelector()
        data_window=WindowSelector()
        opacity = SliderLabel(3)
        opacity.setRange(0,1)
        opacity.setValue(.5)
        show = ShowCheckbox(opacity)
        show.setChecked(True)
        self.items.append({'name':'background_window','string':'Background window','object':background_window})
        self.items.append({'name':'data_window','string':'Data window','object':data_window})
        self.items.append({'name':'opacity','string':'Opacity','object':opacity})
        self.items.append({'name':'show','string':'Show','object':show})
        super().gui()
    def __call__(self, background_window, data_window, opacity,show,keepSourceWindow=False):
        if background_window is None or data_window is None:
            return
        w=data_window
        if show:
            if hasattr(w, 'bgItem') and w.bgItem is not None:
                w.bgItem.hist_luttt.hide()
                w.imageview.ui.gridLayout.removeWidget(w.bgItem.hist_luttt)
                w.imageview.view.removeItem(w.bgItem)
            bgItem = pg.ImageItem(background_window.imageview.imageItem.image)
            bgItem.setOpacity(opacity)
            w.imageview.view.addItem(bgItem)
            bgItem.hist_luttt = pg.HistogramLUTWidget()
            bgItem.hist_luttt.setMinimumWidth(110)
            bgItem.hist_luttt.setImageItem(bgItem)
            w.imageview.ui.gridLayout.addWidget(bgItem.hist_luttt, 0, 4, 1, 4)
            w.bgItem = bgItem
        else:
            if hasattr(w, 'bgItem') and w.bgItem is not None:
                w.bgItem.hist_luttt.hide()
                w.imageview.ui.gridLayout.removeWidget(w.bgItem.hist_luttt)
                w.imageview.view.removeItem(w.bgItem)
                w.bgItem.hist_luttt = None
                w.bgItem = None
            return None
    def preview(self):
        background_window=self.getValue('background_window')
        data_window=self.getValue('data_window')
        opacity=self.getValue('opacity')
        show=self.getValue('show')
        self.__call__(background_window,data_window,opacity,show)
        
background=Background()

class Scale_Bar(BaseProcess):
    ''' scale_bar(width_microns, width_pixels, font_size, color, background, location, show=True)

    Parameters:
        | width_microns (float)
        | width_pixels (float)
        | font_size (int)
        | color ['Black','White'] (string)
        | background  ['Black','White', 'None'] (string)
        | location ['Lower Right','Lower Left','Top Right','Top Left'] (string)
        | show (bool)
    '''
    
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        w=g.currentWindow
        width_microns=QtWidgets.QDoubleSpinBox()
        
        width_pixels=QtWidgets.QSpinBox()
        width_pixels.setRange(.001,1000000)
        width_pixels.setRange(1,w.mx)
        
        font_size=QtWidgets.QSpinBox()
        
        color=ComboBox()
        color.addItem("White")
        color.addItem("Black")
        background=ComboBox()
        background.addItem('None')
        background.addItem('Black')
        background.addItem('White')
        location=ComboBox()
        location.addItem('Lower Right')
        location.addItem('Lower Left')
        location.addItem('Top Right')
        location.addItem('Top Left')
        show=CheckBox()
        if hasattr(w,'scaleBarLabel') and w.scaleBarLabel is not None: #if the scaleBarLabel already exists
            props=w.scaleBarLabel.flika_properties
            width_microns.setValue(props['width_microns'])
            width_pixels.setValue(props['width_pixels'])
            font_size.setValue(props['font_size'])
            color.setCurrentIndex(color.findText(props['color']))
            background.setCurrentIndex(background.findText(props['background']))
            location.setCurrentIndex(location.findText(props['location']))
        else:
            font_size.setValue(12)
            width_pixels.setValue(int(w.mx/8))
            width_microns.setValue(1)
            
        show.setChecked(True) 
        self.items.append({'name':'width_microns','string':'Width of bar in microns','object':width_microns})
        self.items.append({'name':'width_pixels','string':'Width of bar in pixels','object':width_pixels})
        self.items.append({'name':'font_size','string':'Font size','object':font_size})
        self.items.append({'name':'color','string':'Color','object':color})
        self.items.append({'name':'background','string':'Background','object':background})
        self.items.append({'name':'location','string':'Location','object':location})
        self.items.append({'name':'show','string':'Show','object':show})
        
        super().gui()
        self.preview()
    def __call__(self,width_microns, width_pixels, font_size, color, background,location,show=True,keepSourceWindow=None):
        w=g.currentWindow
        if show:
            if hasattr(w,'scaleBarLabel') and w.scaleBarLabel is not None:
                w.imageview.view.removeItem(w.scaleBarLabel.bar)
                w.imageview.view.removeItem(w.scaleBarLabel)
                w.imageview.view.sigResized.disconnect(self.updateBar)
            if location=='Top Left':
                anchor=(0,0)
                pos=[0,0]
            elif location=='Top Right':
                anchor=(0,0)
                pos=[w.mx,0]
            elif location=='Lower Right':
                anchor=(0,0)
                pos=[w.mx,w.my]
            elif location=='Lower Left':
                anchor=(0,0)
                pos=[0,w.my]
            w.scaleBarLabel= pg.TextItem(anchor=anchor, html="<span style='font-size: {}pt;color:{};background-color:{};'>{} μm</span>".format(font_size, color, background,width_microns))
            w.scaleBarLabel.setPos(pos[0],pos[1])
            w.scaleBarLabel.flika_properties={item['name']:item['value'] for item in self.items}
            w.imageview.view.addItem(w.scaleBarLabel)
            if color=='White':
                color255=[255,255,255,255]
            elif color=='Black':
                color255=[0,0,0,255]
            textRect=w.scaleBarLabel.boundingRect()
            
            if location=='Top Left':
                barPoint=QtCore.QPoint(0, textRect.height())
            elif location=='Top Right':
                barPoint=QtCore.QPoint(-width_pixels, textRect.height())
            elif location=='Lower Right':
                barPoint=QtCore.QPoint(-width_pixels, -textRect.height())
            elif location=='Lower Left':
                barPoint=QtCore.QPoint(0, -textRect.height())
                
            bar = QtWidgets.QGraphicsRectItem(QtCore.QRectF(barPoint, QtCore.QSizeF(width_pixels,int(font_size/3))))
            bar.setPen(pg.mkPen(color255)); bar.setBrush(pg.mkBrush(color255))
            w.imageview.view.addItem(bar)
            #bar.setParentItem(w.scaleBarLabel)
            w.scaleBarLabel.bar=bar
            w.imageview.view.sigResized.connect(self.updateBar)
            self.updateBar()
            
        else:
            if hasattr(w,'scaleBarLabel') and w.scaleBarLabel is not None:
                w.imageview.view.removeItem(w.scaleBarLabel.bar)
                w.imageview.view.removeItem(w.scaleBarLabel)
                w.scaleBarLabel=None
                w.imageview.view.sigResized.disconnect(self.updateBar)
        return None
        
    def updateBar(self):
        w=g.currentWindow
        width_pixels=self.getValue('width_pixels')
        location=self.getValue('location')
        view = w.imageview.view
        textRect=w.scaleBarLabel.boundingRect()
        textWidth=textRect.width()*view.viewPixelSize()[0]
        textHeight=textRect.height()*view.viewPixelSize()[1]
        
        if location=='Top Left':
            barPoint=QtCore.QPoint(0, 1.3*textHeight)
            w.scaleBarLabel.setPos(QtCore.QPointF(width_pixels/2-textWidth/2,0))
        elif location=='Top Right':
            barPoint=QtCore.QPoint(w.mx-width_pixels, 1.3*textHeight)
            w.scaleBarLabel.setPos(QtCore.QPointF(w.mx-width_pixels/2-textWidth/2,0))
        elif location=='Lower Right':
            barPoint=QtCore.QPoint(w.mx-width_pixels, w.my-1.3*textHeight)
            w.scaleBarLabel.setPos(QtCore.QPointF(w.mx-width_pixels/2-textWidth/2,w.my-textHeight))
        elif location=='Lower Left':
            barPoint=QtCore.QPoint(0, w.my-1.3*textHeight)
            w.scaleBarLabel.setPos(QtCore.QPointF(QtCore.QPointF(width_pixels/2-textWidth/2,w.my-textHeight)))
        w.scaleBarLabel.bar.setRect(QtCore.QRectF(barPoint, QtCore.QSizeF(width_pixels,textHeight/4)))
        
    def preview(self):
        width_microns=self.getValue('width_microns')
        width_pixels=self.getValue('width_pixels')
        font_size=self.getValue('font_size')
        color=self.getValue('color')
        background=self.getValue('background')
        location=self.getValue('location')
        show=self.getValue('show')
        self.__call__(width_microns, width_pixels, font_size, color, background, location, show)
scale_bar=Scale_Bar()










# -*- coding: utf-8 -*-
import os
import time
import numpy as np
from qtpy import QtCore, QtGui, QtWidgets
import pyqtgraph as pg
from . import global_vars as g
from .utils.misc import save_file_gui

pg.setConfigOptions(useWeave=False)


class TraceFig(QtWidgets.QWidget):
    """Pyqtgraph PlotWidget with frame range selector. Display average trace of ROIs and updates in realtime.
    
    Attributes:
        p1 (pg.PlotWidget): top plot widget displaying selected region of traces
        p2 (pg.PlotWidget): bottom plot widget displaying entire traces with region selection
        rois ([dict]): list of roi information dicts for reference

    Signals:
        :indexChanged(int): emits the index of the mouse when the user hovers over the top plotWidget
        :finishedDrawingSignal(): emits when the bottom ROI is finished updating
        :keyPressSignal(QtCore.QEvent): emits when the traceWindow is selected and a key is pressed
        :partialThreadUpdatedSignal(): emits when the top plot widget is updated
    """
    indexChanged=QtCore.Signal(int)
    finishedDrawingSignal=QtCore.Signal()
    keyPressSignal=QtCore.Signal(QtCore.QEvent)
    partialThreadUpdatedSignal = QtCore.Signal()
    name = "Trace Widget"

    def __init__(self):
        super(TraceFig,self).__init__()
        g.traceWindows.append(self)
        self.setCurrentTraceWindow()
        if 'tracefig_settings' in g.settings and 'coords' in g.settings['tracefig_settings']:
            self.setGeometry(QtCore.QRect(*g.settings['tracefig_settings']['coords']))
        else:
            self.setGeometry(QtCore.QRect(355, 30, 1219, 148))
        self.setWindowTitle('flika')
        self.l = QtWidgets.QVBoxLayout()
        self.l.setContentsMargins(0,0,0,0)
        self.setLayout(self.l)
        self.p1=pg.PlotWidget()
        self.p2=pg.PlotWidget()
        self.p1.getPlotItem().axes['left']['item'].setGrid(100) #this makes faint horizontal lines
        self.p2.setMaximumHeight(50)
        self.export_button = QtWidgets.QPushButton("Export")
        self.export_button.setMaximumWidth(100)
        self.export_button.clicked.connect(self.export_gui)
        self.l.addWidget(self.p1, 1)
        self.l.addWidget(self.p2, 1)
        self.l.addWidget(self.export_button, 0)
    
        self.region = pg.LinearRegionItem()         # Add the LinearRegionItem to the ViewBox, but tell the ViewBox to exclude this item when doing auto-range calculations.
        self.region.setZValue(10)
        self.p2.plotItem.addItem(self.region, ignoreBounds=True)
        self.p1.setAutoVisible(y=True)
        self.rois=[] # roi in this list is a dict: {roi, p1trace,p2trace, sigproxy}
        self.redrawPartialThread = None
        self.vb = self.p1.plotItem.getViewBox()
        
        self.proxy = pg.SignalProxy(self.p1.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved)
        self.p2.plotItem.vb.mouseDragEvent=self.mouseDragEvent2
        
        self.region.sigRegionChanged.connect(self.update)
        self.p1.plotItem.sigRangeChanged.connect(self.updateRegion)
        self.region.setRegion([0, 200])

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.p1.update)
        self.timer.start()

        from .process.measure import measure
        self.measure=measure
        self.p1.scene().sigMouseClicked.connect(self.measure.pointclicked)
        self.p1.scene().sigMouseClicked.connect(self.setCurrentTraceWindow)
        self.resizeEvent = self.onResize
        self.moveEvent = self.onMove
        
        if 'tracefig_settings' not in g.settings:
            g.settings['tracefig_settings']=dict()
            try:
                g.settings['tracefig_settings']['coords']=self.geometry().getRect()
            except Exception as e:
                g.alert(e)
        self.show()
        
    def onResize(self,event):
        g.settings['tracefig_settings']['coords']=self.geometry().getRect()

    def onMove(self,event):
        g.settings['tracefig_settings']['coords']=self.geometry().getRect()
        
    def setCurrentTraceWindow(self):
        if g.currentTrace is not None:
            g.currentTrace.setStyleSheet("border:1px solid rgb(0, 0, 0); ")
        self.setStyleSheet("border:1px solid rgb(0, 255, 0); ")
        g.currentTrace = self

    def mouseDragEvent2(self,ev):
        ev.ignore() # prevent anything from happening

    def mouseDragEvent1(self,ev):
        ev.ignore() # prevent anything from happening

    def keyPressEvent(self,ev):
        self.keyPressSignal.emit(ev)

    def closeEvent(self, event):
        while len(self.rois) > 0:
            self.removeROI(0)
        try:
            self.p1.scene().sigMouseClicked.disconnect(self.measure.pointclicked)
            self.p1.scene().sigMouseClicked.disconnect(self.setCurrentTraceWindow)
        except:
            pass
        if self in g.traceWindows:
            g.traceWindows.remove(self)
        g.currentTrace = None
        event.accept() # let the window close

    def update(self):
        self.region.setZValue(10)
        minX, maxX = self.region.getRegion()
        self.p1.plotItem.setXRange(minX, maxX, padding=0, update=False)    
        self.p1.plotItem.axes['bottom']['item'].setRange(minX,maxX)

    def updateRegion(self,window, viewRange):
        rgn = viewRange[0]
        self.region.setRegion(rgn)

    def getBounds(self):
        bounds=self.region.getRegion()
        bounds=[int(np.floor(bounds[0])),int(np.ceil(bounds[1]))+1]
        return bounds

    def mouseMoved(self,evt):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            pass
        else:
            pos = evt[0]  ## using signal proxy turns original arguments into a tuple
            if self.p1.plotItem.sceneBoundingRect().contains(pos):
                mousePoint = self.vb.mapSceneToView(pos)
                index = int(mousePoint.x())
                if index >= 0:
                    #self.label.setText("<span style='font-size: 12pt'>frame={0}</span>".format(index))
                    self.indexChanged.emit(index)
                    g.m.statusBar().showMessage('frame {}    y={}'.format(index,mousePoint.y()))
    
    def get_roi_index(self,roi):
        return [r['roi'] for r in self.rois].index(roi)
        
    def alert(self, msg):
        #print(msg)
        pass

    def translated(self,roi):
        index=self.get_roi_index(roi)
        self.rois[index]['toBeRedrawn']=True
        if self.redrawPartialThread is None or self.redrawPartialThread.isFinished():
            self.alert('Launching redrawPartialThread')
            self.redrawPartialThread = RedrawPartialThread(self)
            self.redrawPartialThread.alert.connect(self.alert)
            self.redrawPartialThread.start()
            self.redrawPartialThread.updated.connect(self.partialThreadUpdatedSignal.emit)
            
    def translateFinished(self, roi):
        roi_index = self.get_roi_index(roi)
        if self.redrawPartialThread is not None and self.redrawPartialThread.isRunning():
            self.redrawPartialThread.quit_loop = True
            #self.redrawPartialThread.finished_sig.emit() #tell the thread to finish
            #loop = QtCore.QEventLoop()
            #self.redrawPartialThread.finished.connect(loop.quit)
            #loop.exec_()# This blocks until the "finished" signal is emitted
        trace=roi.getTrace()
        self.update_trace_full(roi_index, trace)

    def update_trace_full(self, roi_index, trace):
        pen=QtGui.QPen(self.rois[roi_index]['roi'].pen)
        self.rois[roi_index]['p1trace'].setData(trace,pen=pen)
        self.rois[roi_index]['p2trace'].setData(trace,pen=pen)
        self.finishedDrawingSignal.emit()
        
    def addROI(self,roi):
        if self.hasROI(roi):
            return
        trace=roi.getTrace()
        if trace is None:
            raise InvalidTraceException()
        pen=QtGui.QPen(roi.pen)
        pen.setWidth(0)
        if len(trace)==1:
            p1trace=self.p1.plot(trace, pen=None, symbol='o')
            p2trace=self.p2.plot(trace, pen=None, symbol='o')
        else:
            p1trace=self.p1.plot(trace, pen=pen)
            p2trace=self.p2.plot(trace, pen=pen) 
        
        roi.sigRegionChanged.connect(self.translated)
        roi.sigRegionChangeFinished.connect(self.translateFinished)

        if len(self.rois)==0:
            self.region.setRegion([0, len(trace)-1])
        self.rois.append(dict({'roi':roi,'p1trace':p1trace,'p2trace':p2trace,'toBeRedrawn':False,'toBeRedrawnFull':False}))

    def removeROI(self,roi):
        from .roi import ROI_Base
        if isinstance(roi, ROI_Base):
            index=[r['roi'] for r in self.rois].index(roi) #this is the index of the roi in self.rois
        elif isinstance(roi, int):
            index = roi
        else:
            g.alert("Failed to remove roi {}".format(roi))
            return
        self.p1.removeItem(self.rois[index]['p1trace'])
        self.p2.removeItem(self.rois[index]['p2trace'])
        self.rois[index]['roi'].traceWindow = None
        try:
            self.rois[index]['roi'].resetSignals()
        except:
            pass
        del self.rois[index]
        if len(self.rois)==0:
            self.close()

    def hasROI(self,roi):
        return roi in [r['roi'] for r in self.rois] #return True if roi is plotted

    def export_gui(self):
        filename = g.settings['filename']
        directory = os.path.dirname(filename)
        if filename is not None:
            filename = save_file_gui('Save Traces', directory, '*.txt')
        else:
            filename = save_file_gui('Save Traces', '', '*.txt')
        if filename == '':
            return False
        else:
            self.export(filename)

    def export(self,filename):
        ''' This function saves out all the traces in the tracefig to a file specified by the argument 'filename'.
        The output file is a tab seperated ascii file where each column is a trace.  
        Traces are saved in the order they were added to the plot.
        
        '''
        g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
        traces=[]
        for roi in self.rois:
            traces.append(roi['roi'].getTrace())
        traces.insert(0,np.arange(len(traces[0])))
        traces=np.array(traces).T
        np.savetxt(filename,traces,delimiter='\t',fmt='%10f')
        g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
        
def roiPlot(roi):
    '''
    returns tracefig that is used to plot roi
    '''
    if g.settings['multipleTraceWindows'] or g.currentTrace is None:
        win = TraceFig()
    else:
        win = g.currentTrace
    try:
        win.addROI(roi)
    except InvalidTraceException:
        if len(win.rois) == 0:
            win.close()
            return None
    return win


class RedrawPartialThread(QtCore.QThread):
    finished=QtCore.Signal() #this announces that the thread has finished
    finished_sig=QtCore.Signal() #This tells the thread to finish
    alert = QtCore.Signal(str)
    updated = QtCore.Signal() #This signal is emitted after each redraw

    def __init__(self,tracefig):
        QtCore.QThread.__init__(self)
        self.tracefig=tracefig
        self.redrawCompleted=True
        self.quit_loop=False
        
    def run(self):
        self.finished_sig.connect(self.request_quit_loop)
        while self.quit_loop is False:
            time.sleep(.05)
            self.redraw()
            self.updated.emit()
        self.alert.emit("Finished Redraw")
        self.finished.emit()
        
    def request_quit_loop(self):
        self.quit_loop=True
        
    def redraw(self):
        if self.redrawCompleted is False:
            self.alert.emit("Redraw hasn't finished")
            pass
        else:
            self.alert.emit("Redrawing")
            self.redrawCompleted=False
            idxs=[]
            for i in np.arange(len(self.tracefig.rois)):
                if self.tracefig.rois[i]['toBeRedrawn']:
                    self.tracefig.rois[i]['toBeRedrawn']=False
                    idxs.append(i)
            traces=[]
            bounds=self.tracefig.getBounds()
            bounds = [max(0, bounds[0]), bounds[1]]
            for i in idxs:
                roi=self.tracefig.rois[i]['roi']
                trace=roi.getTrace(bounds)
                traces.append(trace)
            for i, roi_index in enumerate(idxs):
                trace=traces[i] #This function can sometimes take a long time.  
                pen=QtGui.QPen(self.tracefig.rois[roi_index]['roi'].pen)
                bb=self.tracefig.getBounds()
                curve=self.tracefig.rois[roi_index]['p1trace']
                newtrace=curve.getData()[1]
                if bb[0]<0: bb[0]=0
                if bb[1]>len(newtrace): bb[1]=len(newtrace)
                if bb[1]<0 or bb[0]>len(newtrace):
                    return
                newtrace[bb[0]:bb[1]]=trace
                curve.setData(newtrace,pen=pen)
                self.alert.emit("CURVE {} redrawn".format(roi_index))
                QtWidgets.qApp.processEvents()
            self.redrawCompleted=True

class InvalidTraceException(Exception):
    pass






import sys, os
import optparse

from ..process import *
from .. import global_vars as g
from ..window import Window
from ..roi import makeROI
import numpy as np
import time
import pytest
import warnings
warnings.filterwarnings("ignore")


g.settings['multiprocessing'] = False
zproject.gui()
obj = [i for i in zproject.items if i['name'] == 'projection_type'][0]['object']
ZPROJECTS = [obj.itemText(i) for i in range(obj.count())]
zproject.ui.close()

image_calculator.gui()
obj = [i for i in image_calculator.items if i['name'] == 'operation'][0]['object']
OPERANDS = [obj.itemText(i) for i in range(obj.count())]
image_calculator.ui.close()



@pytest.mark.parametrize("img", [
    (np.random.randint(2, size=[10, 20, 20]).astype("uint8")),
    (np.random.randint(2, size=[10, 20, 20]).astype("uint16")),
    (np.random.randint(2, size=[10, 20, 20]).astype("uint32")),
    (np.random.randint(2, size=[10, 20, 20]).astype("uint64")),
    (np.random.randint(2, size=[10, 20, 20]).astype("int8")),
    (np.random.randint(2, size=[10, 20, 20]).astype("int16")),
    (np.random.randint(2, size=[10, 20, 20]).astype("int32")),
    (np.random.randint(2, size=[10, 20, 20]).astype("int64")),
    (np.random.randint(2, size=[10, 20, 20]).astype("float16")),
    (np.random.randint(2, size=[10, 20, 20]).astype("float32")),
    (np.random.randint(2, size=[10, 20, 20]).astype("float64")),
    ((np.random.random([10, 20, 20])*10).astype("uint8")),
    ((np.random.random([10, 20, 20])*10).astype("uint16")),
    ((np.random.random([10, 20, 20])*10).astype("uint32")),
    ((np.random.random([10, 20, 20])*10).astype("uint64")),
    ((np.random.random([10, 20, 20])*10).astype("int8")),
    ((np.random.random([10, 20, 20])*10).astype("int16")),
    ((np.random.random([10, 20, 20])*10).astype("int32")),
    ((np.random.random([10, 20, 20])*10).astype("int64")),
    (np.random.random([10, 20, 20]).astype('float16')),
    (np.random.random([10, 20, 20]).astype('float32')),
    (np.random.random([10, 20, 20]).astype('float64')),
   	((np.random.random([10, 20, 20, 3])*10).astype("uint8")),
    ((np.random.random([10, 20, 20, 3])*10).astype("uint16")),
    ((np.random.random([10, 20, 20, 3])*10).astype("uint32")),
    ((np.random.random([10, 20, 20, 3])*10).astype("uint64")),
    ((np.random.random([10, 20, 20, 3])*10).astype("int8")),
    ((np.random.random([10, 20, 20, 3])*10).astype("int16")),
    ((np.random.random([10, 20, 20, 3])*10).astype("int32")),
    ((np.random.random([10, 20, 20, 3])*10).astype("int64")),
    (np.random.random([10, 20, 20, 3]).astype('float16')),
    (np.random.random([10, 20, 20, 3]).astype('float32')),
    (np.random.random([10, 20, 20, 3]).astype('float64')),
   	((np.random.random([20, 20])*10).astype("uint8")),
    ((np.random.random([20, 20])*10).astype("uint16")),
    ((np.random.random([20, 20])*10).astype("uint32")),
    ((np.random.random([20, 20])*10).astype("uint64")),
    ((np.random.random([20, 20])*10).astype("int8")),
    ((np.random.random([20, 20])*10).astype("int16")),
    ((np.random.random([20, 20])*10).astype("int32")),
    ((np.random.random([20, 20])*10).astype("int64")),
    (np.random.random([20, 20]).astype('float16')),
    (np.random.random([20, 20]).astype('float32')),
    (np.random.random([20, 20]).astype('float64')),
   	((np.random.random([20, 20, 3])*10).astype("uint8")),
    ((np.random.random([20, 20, 3])*10).astype("uint16")),
    ((np.random.random([20, 20, 3])*10).astype("uint32")),
    ((np.random.random([20, 20, 3])*10).astype("uint64")),
    ((np.random.random([20, 20, 3])*10).astype("int8")),
    ((np.random.random([20, 20, 3])*10).astype("int16")),
    ((np.random.random([20, 20, 3])*10).astype("int32")),
    ((np.random.random([20, 20, 3])*10).astype("int64")),
    (np.random.random([20, 20, 3]).astype('float16')),
    (np.random.random([20, 20, 3]).astype('float32')),
    (np.random.random([20, 20, 3]).astype('float64')),
])
class ProcessTest:
	def teardown_method(self):
		from .conftest import fa
		fa().clear()


class TestBinary(ProcessTest):
	
	def test_threshold(self, img, fa):
		w1 = Window(img)
		w = threshold(.5)
		

	def test_adaptive_threshold(self, img, fa):
		w1 = Window(img)
		w = adaptive_threshold(.5, 3)
		

	def test_canny_edge_detector(self, img):
		if img.ndim == 4:
			return
		w1 = Window(img)
		w = canny_edge_detector(.5)
		
	
	def test_binary_dilation(self, img):
		if img.ndim == 4 or not ((img==0)|(img==1)).all():
			return
		w1 = Window(img)
		w = binary_dilation(2, 3, 1)
		
	
	def test_binary_erosion(self, img):
		if img.ndim == 4 or not ((img==0)|(img==1)).all():
			return
		w1 = Window(img)
		w = binary_erosion(2, 3, 1)
		

	def test_generate_rois(self, img):
		if img.ndim == 4 or not ((img==0)|(img==1)).all():
			return
		w1 = Window(img)
		w = generate_rois(.5, 10)
		

	def test_remove_small_blobs(self, img):
		if img.ndim == 4 or not ((img==0)|(img==1)).all():
			return
		w1 = Window(img)
		w = threshold(.5)
		


class TestFilters(ProcessTest):
	def test_gaussian_blur(self, img):
		w1 = Window(img)
		w = gaussian_blur(.5)
		

	def test_butterworth_filter(self, img):
		w1 = Window(img)
		w = butterworth_filter(1, .2, .6)
		

	def test_mean_filter(self, img):
		w1 = Window(img)
		w = mean_filter(5)
		

	def test_median_filter(self, img):
		w1 = Window(img)
		w = median_filter(5)
		

	def test_fourier_filter(self, img):
		w1 = Window(img)
		w = fourier_filter(3, .2, .6, False)
		

	def test_difference_filter(self, img):
		w1 = Window(img)
		w = difference_filter()
		

	def test_boxcar_differential_filter(self, img):
		w1 = Window(img)
		w = boxcar_differential_filter(2, 3)
		

	def test_wavelet_filter(self, img):
		w1 = Window(img)
		w = wavelet_filter(2, 3)
		

	def test_bilateral_filter(self, img):
		w1 = Window(img)
		w = bilateral_filter(True, 30, 10, .05, 100) # soft filter
		w2 = bilateral_filter(False, 30, 10, .05, 100) # hard filter
		

class TestMath(ProcessTest):
	def test_subtract(self, img):
		w1 = Window(img)
		subtract(2)
		

	def test_subtract_trace(self, img):
		w1 = Window(img)
		roi1 = makeROI('rectangle', [[3, 3], [5, 6]])
		tr = roi1.plot()
		if tr:
			subtract_trace()
		

	def test_divide_trace(self, img):
		w1 = Window(img)
		roi1 = makeROI('rectangle', [[3, 3], [5, 6]])
		tr = roi1.plot()
		if tr:
			divide_trace()
		

	def test_multiply(self, img):
		w1 = Window(img)
		multiply(2.4)
		

	def test_power(self, img):
		w1 = Window(img)
		power(2)
		

	def test_ratio(self, img):
		w1 = Window(img)
		ratio(2, 6, 'average')
		ratio(2, 6, 'standard deviation')
		

	def test_absolute_value(self, img):
		w1 = Window(img)
		absolute_value()
		

class TestOverlay(ProcessTest):
	def test_time_stamp(self, img):
		w1 = Window(img)
		time_stamp(2)
		

	def test_background(self, img):
		w1 = Window(img)
		w2 = Window(img/2)
		background(w1, w2, .5, True)
		

	def test_scale_bar(self, img):
		w1 = Window(img)
		scale_bar.gui()
		scale_bar(30, 5, 12, 'White', 'None','Lower Left')
		

class TestColor(ProcessTest):
	def test_split_channels(self, img):
		if img.ndim == 4 or (img.ndim == 3 and img.shape[2] == 3):
			w1 = Window(img)
			split_channels()
		

class TestROIProcess(ProcessTest):
	def test_set_value(self, img):
		w1 = Window(img)
		roi = makeROI('rectangle', [[3, 3], [4, 5]])
		set_value(2, 2, 5)
		

class TestStacks(ProcessTest):
	def test_deinterleave(self, img):
		w1 = Window(img)
		deinterleave(2)
		

	def test_trim(self, img):
		w1 = Window(img)
		trim(2, 6, 2)
		

	def test_zproject(self, img):
		w1 = Window(img)
		
		for i in ZPROJECTS:
			w3 = zproject(2, 6, i, True)
			if isinstance(w3, Window):
				w3.close()
			w1.setAsCurrentWindow()
		

	def test_image_calculator(self, img):
		w1 = Window(img)
		w2 = Window(img)
		for i in OPERANDS:
			w3 = image_calculator(w1, w2, i, True)
			if isinstance(w3, Window):
				w3.close()
		

	def test_pixel_binning(self, img):
		w1 = Window(img)
		pixel_binning(2)
		

	def test_frame_binning(self, img):
		w1 = Window(img)
		frame_binning(2)
		

	def test_resize(self, img):
		w1 = Window(img)
		resize(2)
		



from ..app.plugin_manager import PluginManager, load_local_plugins, plugin_list
from ..app.script_editor import ScriptEditor
from ..window import Window
import numpy as np

class TestPluginManager():
	def setup_method(self, method):
		PluginManager.show()
		load_local_plugins()
		
	def teardown_method(self, method):
		PluginManager.close()

	def test_local_plugins(self):
		local = set(plugin_list.keys())
		plugins = set(PluginManager.plugins.keys())

		assert (local & plugins) == local, "Local plugin list not loaded correctly"

	def test_install_plugin(self):
		load_local_plugins()
		plugin_name = 'Detect Puffs'
		if PluginManager.plugins[plugin_name].installed:
			return
		PluginManager.refresh_online_plugins()
		i = 0
		while PluginManager.plugins[plugin_name].url == None:
			i += 1
			if i == 1000:
				return
			continue
		PluginManager.downloadPlugin(plugin_name)
		assert PluginManager.plugins[plugin_name].menu is not None and PluginManager.plugins[plugin_name].installed, "Plugin install"
		PluginManager.removePlugin(plugin_name)
		assert PluginManager.plugins[plugin_name].menu is None and not PluginManager.plugins[plugin_name].installed, "Plugin uninstall"


class TestScriptEditor():
	def setup_method(self, method):
		ScriptEditor.show()

	def teardown_method(self, method):
		ScriptEditor.close()

	def test_from_window(self):
		w1 = Window(np.random.random([10, 20, 20]))
		from ..process import threshold
		w2 = threshold(.5)
		ScriptEditor.gui.actionFrom_Window.trigger()
		text = str(ScriptEditor.gui.currentTab().toPlainText())
		assert text == "threshold(value=0.5, darkBackground=False, keepSourceWindow=False)", "From window command not expected"


# -*- coding: utf-8 -*-
import pyqtgraph as pg
import pyqtgraph.exporters
import time
import os.path
import numpy as np
from qtpy import uic, QtGui, QtCore, QtWidgets
import codecs
import shutil, subprocess
import json
import re
import nd2reader
import datetime
import json
import pip
import re

from .. import global_vars as g
from ..app.script_editor import ScriptEditor
from .BaseProcess import BaseDialog
from ..window import Window
from ..utils.misc import open_file_gui, save_file_gui
from ..utils.io import tifffile

__all__ = ['save_file', 'save_points', 'save_movie_gui', 'open_file', 'open_file_from_gui', 'open_points', 'close']

########################################################################################################################
######################                  SAVING FILES                                         ###########################
########################################################################################################################



def save_file(filename=None):
    """
    Save the image in the currentWindow to a .tif file.

    Parameters
    ----------
    filename : str
        The image or movie will be saved here.

    """
    if filename is None or filename is False:
        filetypes = '*.tif'
        prompt = 'Save File As Tif'
        filename = save_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    if os.path.dirname(filename) == '':  # if the user didn't specify a directory
        directory = os.path.normpath(os.path.dirname(g.settings['filename']))
        filename = os.path.join(directory, filename)
    g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
    A = g.currentWindow.image
    metadata = g.currentWindow.metadata
    try:
        metadata = json.dumps(metadata, default=JSONhandler)
    except TypeError as e:
        msg = "Error saving metadata.\n{}\nContinuing to save file".format(e)
        g.alert(msg)
    if len(A.shape) == 3:
        A = np.transpose(A, (0, 2, 1))  # This keeps the x and the y the same as in FIJI
    elif len(A.shape) == 2:
        A = np.transpose(A, (1, 0))
    tifffile.imsave(filename, A,
                    description=metadata)  # http://stackoverflow.com/questions/20529187/what-is-the-best-way-to-save-image-metadata-alongside-a-tif-with-python
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
    return filename

def save_points(filename=None):
    if filename is None:
        filetypes = '*.txt'
        prompt = 'Save Points'
        filename = save_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    g.m.statusBar().showMessage('Saving Points in {}'.format(os.path.basename(filename)))
    p_out = []
    p_in = g.currentWindow.scatterPoints
    for t in np.arange(len(p_in)):
        for p in p_in[t]:
            p_out.append(np.array([t, p[0], p[1]]))
    p_out = np.array(p_out)
    np.savetxt(filename, p_out)
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
    return filename


def save_movie_gui():
    rateSpin = pg.SpinBox(value=50, bounds=[1, 1000], suffix='fps', int=True, step=1)
    rateDialog = BaseDialog([{'string': 'Framerate', 'object': rateSpin}], 'Save Movie', 'Set the framerate')
    rateDialog.accepted.connect(lambda: save_movie(rateSpin.value()))
    g.dialogs.append(rateDialog)
    rateDialog.show()


def save_movie(rate, filename=None):
    """save_movie(rate, filename)
    Saves the currentWindow video as a .mp4 movie by joining .jpg frames together

    Parameters:
        | rate (int) -- framerate
        | filename (str) -- Address to save the movie to, with .mp4

    Notes:
        | Once you've exported all of the frames you wanted, open a command line and run the following:
        |   ffmpeg -r 100 -i %03d.jpg output.mp4
        | -r: framerate
        | -i: input files.
        | %03d: The files have to be numbered 001.jpg, 002.jpg... etc.
    """


    ## Check if ffmpeg is installed
    if os.name == 'nt':  # If we are running windows
        try:
            subprocess.call(["ffmpeg"])
        except FileNotFoundError as e:
            if e.errno == os.errno.ENOENT:
                # handle file not found error.
                # I used http://ffmpeg.org/releases/ffmpeg-2.8.4.tar.bz2 originally
                g.alert("The program FFmpeg is required to export movies. \
                \n\nFor instructions on how to install, go here: http://www.wikihow.com/Install-FFmpeg-on-Windows")
                return None
            else:
                # Something else went wrong while trying to run `wget`
                raise

    filetypes = "Movies (*.mp4)"
    prompt = "Save movie to .mp4 file"
    filename = save_file_gui(prompt, filetypes=filetypes)
    if filename is None:
        return None

    win = g.currentWindow
    A = win.image
    if len(A.shape) < 3:
        g.alert('Movie not the right shape for saving.')
        return None
    try:
        exporter = pg.exporters.ImageExporter(win.imageview.view)
    except TypeError:
        exporter = pg.exporters.ImageExporter.ImageExporter(win.imageview.view)

    nFrames = len(A)
    tmpdir = os.path.join(os.path.dirname(g.settings.settings_file), 'tmp')
    if os.path.isdir(tmpdir):
        shutil.rmtree(tmpdir)
    os.mkdir(tmpdir)
    win.top_left_label.hide()
    for i in np.arange(0, nFrames):
        win.setIndex(i)
        exporter.export(os.path.join(tmpdir, '{:03}.jpg'.format(i)))
        QtWidgets.qApp.processEvents()
    win.top_left_label.show()
    olddir = os.getcwd()
    os.chdir(tmpdir)
    subprocess.call(
        ['ffmpeg', '-r', '%d' % rate, '-i', '%03d.jpg', '-vf', 'scale=trunc(iw/2)*2:trunc(ih/2)*2', 'output.mp4'])
    os.rename('output.mp4', filename)
    os.chdir(olddir)
    g.m.statusBar().showMessage('Successfully saved movie as {}.'.format(os.path.basename(filename)))










########################################################################################################################
######################                         OPENING FILES                                 ###########################
########################################################################################################################


def open_file_from_gui():
    open_file(None, True)


def open_file(filename=None, from_gui=False):
    """ open_file(filename=None)
    Opens an image or movie file (.tif, .stk, .nd2) into a newWindow.

    Parameters:
        | filename (str) -- Address of file to open. If no filename is provided, the last opened file is used.
    Returns:
        newWindow
    """
    if filename is None:
        if from_gui:
            filetypes = 'Image Files (*.tif *.stk *.tiff *.nd2);;All Files (*.*)'
            prompt = 'Open File'
            filename = open_file_gui(prompt, filetypes=filetypes)
            if filename is None:
                return None
        else:
            filename = g.settings['filename']
            if filename is None:
                g.alert('No filename selected')
                return None
    g.m.statusBar().showMessage('Loading {}'.format(os.path.basename(filename)))
    t = time.time()
    metadata = dict()
    ext = os.path.splitext(filename)[1]
    if ext in ['.tif', '.stk', '.tiff', '.ome']:
        try:
            Tiff = tifffile.TiffFile(filename)
        except Exception as s:
            g.alert("Unable to open {}. {}".format(filename, s))
            return None
        metadata = get_metadata_tiff(Tiff)
        A = Tiff.asarray()
        Tiff.close()
        axes = [tifffile.AXES_LABELS[ax] for ax in Tiff.series[0].axes]
        # print("Original Axes = {}".format(Tiff.series[0].axes)) #sample means RBGA, plane means frame, width means X, height means Y
        try:
            assert len(axes) == len(A.shape)
        except AssertionError:
            msg = 'Tiff could not be loaded because the number of axes in the array does not match the number of axes found by tifffile.py\n'
            msg += "Shape of array: {}\n".format(A.shape)
            msg += "Axes found by tifffile.py: {}\n".format(axes)
            g.alert(msg)
            return None
        if set(axes) == set(['height', 'width']):  # still image in black and white.
            target_axes = ['width', 'height']
        elif set(axes) == set(['height', 'width', 'channel']):  # still image in color.
            target_axes = ['width', 'height', 'channel']
            metadata['is_rgb'] = True
        elif set(axes) == set(['height', 'width', 'sample']):  # still image in color.
            target_axes = ['width', 'height', 'sample']
            metadata['is_rgb'] = True
        elif set(axes) == set(['height', 'width', 'series']):  # movie in black and white
            target_axes = ['series', 'width', 'height']
        elif set(axes) == set(['height', 'width', 'time']):  # movie in black and white
            target_axes = ['time', 'width', 'height']
        elif set(axes) == set(['height', 'width', 'depth']):  # movie in black and white
            target_axes = ['depth', 'width', 'height']
        elif set(axes) == set(['channel', 'time', 'height', 'width']):  # movie in color
            target_axes = ['time', 'width', 'height', 'channel']
            metadata['is_rgb'] = True
        elif set(axes) == set(['sample', 'time', 'height', 'width']):  # movie in color
            target_axes = ['time', 'width', 'height', 'sample']
            metadata['is_rgb'] = True
        perm = get_permutation_tuple(axes, target_axes)
        A = np.transpose(A, perm)
        if target_axes[-1] in ['channel', 'sample', 'series'] and A.shape[-1] == 2:
            B = np.zeros(A.shape[:-1])
            B = np.expand_dims(B, len(B.shape))
            A = np.append(A, B, len(A.shape)-1) # add a column of zeros to the last dimension.


        #if A.ndim == 4 and axes[3] == 'sample' and A.shape[3] == 1:
        #    A = np.squeeze(A)  # this gets rid of the meaningless 4th dimention in .stk files


    elif ext == '.nd2':
        nd2 = nd2reader.ND2Reader(filename)
        axes = nd2.axes
        mx = nd2.metadata['width']
        my = nd2.metadata['height']
        mt = nd2.metadata['total_images_per_channel']
        A = np.zeros((mt, mx, my))
        percent = 0
        for frame in range(mt):
            A[frame] = nd2[frame].T
            if percent < int(100 * float(frame) / mt):
                percent = int(100 * float(frame) / mt)
                g.m.statusBar().showMessage('Loading file {}%'.format(percent))
                QtWidgets.qApp.processEvents()
        metadata = nd2.metadata
    elif ext == '.py':
        ScriptEditor.importScript(filename)
        return
    elif ext == '.whl':
        # first, remove trailing (1) or (2)
        newfilename = re.sub(r' \([^)]*\)', '', filename)
        try:
            os.rename(filename, newfilename)
        except FileExistsError:
            pass
        filename = newfilename
        result = pip.main(['install', filename])
        if result == 0:
            g.alert('Successfully installed {}'.format(filename))
        else:
            g.alert('Install of {} failed'.format(filename))
        return
    else:
        msg = "Could not open.  Filetype for '{}' not recognized".format(filename)
        g.alert(msg)
        if filename in g.settings['recent_files']:
            g.settings['recent_files'].remove(filename)
        # make_recent_menu()
        return
        
    append_recent_file(filename)  # make first in recent file menu
    g.m.statusBar().showMessage('{} successfully loaded ({} s)'.format(os.path.basename(filename), time.time() - t))
    g.settings['filename'] = filename
    commands = ["open_file('{}')".format(filename)]
    newWindow = Window(A, os.path.basename(filename), filename, commands, metadata)
    return newWindow

        
def open_points(filename=None):
    if g.currentWindow is None:
        g.alert('Points cannot be loaded if no window is selected. Open a file and click on a window.')
        return None
    if filename is None:
        filetypes = '*.txt'
        prompt = 'Load Points'
        filename = open_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    g.m.statusBar().showMessage('Loading points from {}'.format(os.path.basename(filename)))
    try:
        pts = np.loadtxt(filename)
    except UnicodeDecodeError:
        g.alert('This points file contains text that cannot be read. No points loaded.')
        return None
    if len(pts) == 0:
        g.alert('This points file is empty. No points loaded.')
        return None
    nCols = pts.shape[1]
    pointSize = g.settings['point_size']
    pointColor = QtGui.QColor(g.settings['point_color'])
    if nCols == 3:
        for pt in pts:
            t = int(pt[0])
            if g.currentWindow.mt == 1:
                t = 0
            g.currentWindow.scatterPoints[t].append([pt[1],pt[2], pointColor, pointSize])
        t = g.currentWindow.currentIndex
        g.currentWindow.scatterPlot.setPoints(pos=g.currentWindow.scatterPoints[t])
    elif nCols == 2:
        t = 0
        for pt in pts:
            g.currentWindow.scatterPoints[t].append([pt[0], pt[1], pointColor, pointSize])
        t = g.currentWindow.currentIndex
        g.currentWindow.scatterPlot.setPoints(pos=g.currentWindow.scatterPoints[t])
    g.m.statusBar().showMessage('Successfully loaded {}'.format(os.path.basename(filename)))

    


    
########################################################################################################################
######################                INTERNAL HELPER FUNCTIONS                              ###########################
########################################################################################################################
def get_permutation_tuple(src, dst):
    """

    Parameters
    ----------
    src (list): The original ordering of the axes in the tiff.
    dst (list): The desired ordering of the axes in the tiff.

    Returns
    -------
    result (tuple): The required permutation so the axes are ordered as desired.
    """
    result = []
    for i in dst:
        result.append(src.index(i))
    result = tuple(result)
    return result

def append_recent_file(fname):
    fname = os.path.abspath(fname)
    if fname in g.settings['recent_files']:
        g.settings['recent_files'].remove(fname)
    if os.path.exists(fname):
        g.settings['recent_files'].append(fname)
        if len(g.settings['recent_files']) > 8:
            g.settings['recent_files'] = g.settings['recent_files'][-8:]
    return fname


def get_metadata_tiff(Tiff):
    metadata = {}
    if hasattr(Tiff[0], 'is_micromanager') and Tiff[0].is_micromanager:
        imagej_tags_unpacked = {}
        if hasattr(Tiff[0],'imagej_tags'):
            imagej_tags = Tiff[0].imagej_tags
            imagej_tags['info']
            imagej_tags_unpacked = json.loads(imagej_tags['info'])
        micromanager_metadata = Tiff[0].tags['micromanager_metadata']
        metadata = {**micromanager_metadata.value, **imagej_tags_unpacked}
        if 'Frames' in metadata and metadata['Frames'] > 1:
            timestamps = [c.tags['micromanager_metadata'].value['ElapsedTime-ms'] for c in Tiff]
            metadata['timestamps'] = timestamps
            metadata['timestamp_units'] = 'ms'
        keys_to_remove = ['NextFrame', 'ImageNumber', 'Frame', 'FrameIndex']
        for key in keys_to_remove:
            metadata.pop(key)
    else:
        try:
            metadata = Tiff[0].image_description
            metadata = txt2dict(metadata)
        except AttributeError:
            metadata = dict()
    metadata['is_rgb'] = Tiff[0].is_rgb
    return metadata



def txt2dict(metadata):
    meta = dict()
    try:
        metadata = json.loads(metadata.decode('utf-8'))
        return metadata
    except ValueError:  # if the metadata isn't in JSON
        pass
    for line in metadata.splitlines():
        line = re.split('[:=]', line.decode())
        if len(line) == 1:
            meta[line[0]] = ''
        else:
            meta[line[0].lstrip().rstrip()] = line[1].lstrip().rstrip()
    return meta


def JSONhandler(obj):
    if isinstance(obj,datetime.datetime):
        return obj.isoformat()
    else:
        json.JSONEncoder().default(obj)


def close(windows=None):
    '''
    Will close a window or a set of windows.

    Values for windows:
        | 'all' (str) -- closes all windows
        | windows (list) - closes each window in the list
        | (Window) - closes individual window
        | (None) - closes current window
    '''
    if isinstance(windows, str):
        if windows == 'all':
            windows = [window for window in g.windows]
            for window in windows:
                window.close()
    elif isinstance(windows,list):
        for window in windows:
            if isinstance(window,Window):
                window.close()
    elif isinstance(windows,Window):
        windows.close()
    elif windows is None:
        if g.currentWindow is not None:
            g.currentWindow.close()


########################################################################################################################
######################             OLD FUNCTIONS THAT MIGHT BE USEFUL SOMEDAY                ###########################
########################################################################################################################
"""

def save_roi_traces(filename):
    g.m.statusBar().showMessage('Saving traces to {}'.format(os.path.basename(filename)))
    to_save = [roi.getTrace() for roi in g.currentWindow.rois]
    np.savetxt(filename, np.transpose(to_save), header='\t'.join(['ROI %d' % i for i in range(len(to_save))]), fmt='%.4f', delimiter='\t', comments='')
    g.settings['filename'] = filename
    g.m.statusBar().showMessage('Successfully saved traces to {}'.format(os.path.basename(filename)))

def load_metadata(filename=None):
    '''This function loads the .txt file corresponding to a file into a dictionary
    The .txt is a file which includes database connection information'''
    meta=dict()
    if filename is None:
        filename=os.path.splitext(g.settings['filename'])[0]+'.txt'
    BOM = codecs.BOM_UTF8.decode('utf8')
    if not os.path.isfile(filename):
        g.alert("'"+filename+"' is not a file.")
        return dict()
    with codecs.open(filename, encoding='utf-8') as f:
        for line in f:
            line = line.lstrip(BOM)
            line=line.split('=')
            meta[line[0].lstrip().rstrip()]=line[1].lstrip().rstrip()
    for k in meta.keys():
        if meta[k].isdigit():
            meta[k]=int(meta[k])
        else:
            try:
                meta[k]=float(meta[k])
            except ValueError:
                pass
    return meta

def save_metadata(meta):
    filename=os.path.splitext(g.settings['filename'])[0]+'.txt'
    f=open(filename, 'w')
    text=''
    for item in meta.items():
        text+="{}={}\n".format(item[0],item[1])
    f.write(text)
    f.close()


def save_current_frame(filename):
    "" save_current_frame(filename)
    Save the current single frame image of the currentWindow to a .tif file.

    Parameters:
        | filename (str) -- Address to save the frame to.
    ""
    if os.path.dirname(filename)=='': #if the user didn't specify a directory
        directory=os.path.normpath(os.path.dirname(g.settings['filename']))
        filename=os.path.join(directory,filename)
    g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
    A=np.average(g.currentWindow.image, 0)#.astype(g.settings['internal_data_type'])
    metadata=json.dumps(g.currentWindow.metadata)
    if len(A.shape)==3:
        A = A[g.currentWindow.currentIndex]
        A=np.transpose(A,(0,2,1)) # This keeps the x and the y the same as in FIJI
    elif len(A.shape)==2:
        A=np.transpose(A,(1,0))
    tifffile.imsave(filename, A, description=metadata) #http://stackoverflow.com/questions/20529187/what-is-the-best-way-to-save-image-metadata-alongside-a-tif-with-python
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))

def make_recent_menu():
    g.m.menuRecent_Files.clear()
    if len(g.settings['recent_files']) == 0:
        no_recent = QtWidgets.QAction("No Recent Files", g.m)
        no_recent.setEnabled(False)
        g.m.menuRecent_Files.addAction(no_recent)
        return
    def openFun(f):
        return lambda: open_file(append_recent_file(f))
    for fname in g.settings['recent_files'][:10]:
        if os.path.exists(fname):
            g.m.menuRecent_Files.addAction(QtWidgets.QAction(fname, g.m, triggered=openFun(fname)))

"""


# -*- coding: utf-8 -*-
import os
import numpy as np
import pyqtgraph as pg
from qtpy import QtWidgets, QtCore, QtGui
from .. import global_vars as g
from .BaseProcess import BaseProcess
from ..utils.misc import save_file_gui
from ..roi import ROI_Base

__all__ = ['measure']


np.set_printoptions(suppress=True)
np.set_printoptions(precision=3)
            
class Measure(BaseProcess):
    """
    Click in the graph to select a point.
    Shift-Click in the graph to select the nearest data point
    
    """
    def __init__(self):
        self.ON=False
        super().__init__()
    def gui(self):
        self.gui_reset()
        self.currentPoint = 0
        active_point = QtWidgets.QLineEdit() #active_point.setEnabled(False)
        second_point = QtWidgets.QLineEdit() #second_point.setEnabled(False)
        difference = QtWidgets.QLineEdit() #difference.setEnabled(False)
        slope = QtWidgets.QLineEdit() #slope.setEnabled(False)
        log = QtWidgets.QPushButton('Log Data')
        log.pressed.connect(self.log)
        newcol=QtWidgets.QPushButton('New Column')
        newcol.pressed.connect(self.newcol)
        self.table = pg.TableWidget()
        self.items.append({'name':'active_point','string':'Active Point: ','object':active_point})
        self.items.append({'name':'second_point','string':'Second Point: ','object':second_point})
        self.items.append({'name':'difference','string':'Difference: ','object':difference})
        self.items.append({'name':'slope','string':'Slope: ','object':slope})
        self.items.append({'name':'log','string':'','object':log})
        self.items.append({'name':'newcol','string':'','object':newcol})
        super().gui()
        self.fig = None
        self.ui.accepted.disconnect()
        self.ui.closeSignal.connect(self.close)
        self.ui.rejected.connect(self.close)
        self.ui.accepted.connect(self.close)
        self.ui.accepted.connect(self.export_gui)
        self.data = [[]]
        self.ON = True
    def log(self):
        point=[self.firstPoint[0],self.firstPoint[1],self.secondPoint[0],self.secondPoint[1],self.difference[0],self.difference[1],self.slope]
        self.data[len(self.data)-1].append(point)
    def newcol(self):
        self.data.append([])

    def clear(self):
        self.currentPoint = 0
        try:
            self.viewbox.removeItem(self.pathitem)
        except:
            pass

    def pointclicked(self,evt, window=None):
        if evt.button() != 1:
            return
        if self.ON is False:
            return
        pos = evt.pos()
        if isinstance(evt.currentItem, (ROI_Base, pg.ROI)):
            pos = evt.currentItem.mapToScene(pos)

        if window != None:
            if window != self.fig:
                self.clear()
                self.fig = window
                self.viewbox = self.fig.imageview.view
                self.pathitem=QtWidgets.QGraphicsPathItem(self.viewbox)
                self.pathitem.setPen(QtGui.QPen(QtCore.Qt.red, 0))
                self.viewbox.addItem(self.pathitem,ignoreBounds=True)
            mousePoint = self.fig.imageview.getImageItem().mapFromScene(pos)
            pos = np.array([mousePoint.y(),mousePoint.x()])
        else:
            if self.fig is not g.currentTrace: #if we created a new tracefig
                self.clear()
                self.fig=g.currentTrace
                self.viewbox=self.fig.vb
                if not g.currentTrace.p1.plotItem.sceneBoundingRect().contains(pos):
                    return
                self.pathitem=QtWidgets.QGraphicsPathItem(self.viewbox)
                self.pathitem.setPen(QtGui.QPen(QtCore.Qt.red, 0))
                self.viewbox.addItem(self.pathitem,ignoreBounds=True)
            mousePoint = self.viewbox.mapSceneToView(pos)
            pos = np.array([mousePoint.x(),mousePoint.y()])
        
        self.update(pos)

    def update(self, point):
        
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            point=self.getNearestPoint(point)
        if self.currentPoint==0:
            self.currentPoint=1
            self.firstPoint=point
            self.ui.items[0]['object'].setText(str(point))
            self.ui.items[1]['object'].setText('')
            self.ui.items[2]['object'].setText('')
            self.ui.items[3]['object'].setText('')
            self.secondPoint=None
            self.draw([self.firstPoint])
        else:
            self.currentPoint=0
            self.secondPoint=point
            self.ui.items[1]['object'].setText(str(point))
            self.difference=self.secondPoint-self.firstPoint
            self.ui.items[2]['object'].setText(str(self.difference))
            if self.difference[0]==0:
                self.slope=np.inf
            else:
                self.slope=self.difference[1]/self.difference[0]
            self.ui.items[3]['object'].setText(str(self.slope))
            self.draw([self.firstPoint,self.secondPoint])
                    
    def draw(self,points):
        if type(self.fig) != type(g.currentTrace):
            points = [p[::-1] for p in points]
        path=QtGui.QPainterPath(QtCore.QPointF(*points[0]))
        for i in np.arange(1,len(points)):
            path.lineTo(QtCore.QPointF(*points[i]))
        self.pathitem.setPath(path)
        
    def close(self):
        if hasattr(self, "pathitem") and self.pathitem.parentWidget() != None:
            self.pathitem.parentWidget().removeItem(self.pathitem) 
        self.ON=False
    def getNearestPoint(self,point):
        if hasattr(self.fig, 'imageview'):
            return point
        roi=g.currentTrace.rois[0]
        d=roi['p2trace'].getData()
        index=np.abs(d[0]-point[0]).argmin()
        x=d[0][index]
        ys=[]
        for roi in g.currentTrace.rois:
            d=roi['p2trace'].getData()
            ys.append(d[1][index])
        roi_idx=np.argmin(abs(ys - point[1]))
        y=ys[roi_idx]
        return np.array([x,y])
    
    def export_gui(self):
        filename = g.settings['filename']
        directory = os.path.dirname(filename)
        if filename is not None:
            filename = save_file_gui('Save Measurements', directory, '*.txt')
        else:
            filename = save_file_gui('Save Measurements', None, '*.txt')
        filename = str(filename)
        if filename == '':
            return False
        else:
            self.export(filename)
            
    def export(self,filename):
        ''' This function saves out all the traces in the tracefig to a file specified by the argument 'filename'.
        The output file is a tab seperated ascii file where each column is a trace.  
        Traces are saved in the order they were added to the plot.
        
        '''
        g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
        d=self.data
        output=''
        maxj=np.max([len(j) for j in d])
        header=['x1' ,'y1','x2','y2','x difference','y difference' , 'slope']
        for i in np.arange(len(d)):
            for k in np.arange(7):
                output+=header[k]+'\t'
            output+='\t'
        output+='\n'
        for j in np.arange(maxj): #this loops through each row of data
            for i in np.arange(len(d)): #this loops through each mega-column, each of which is composed of 7 mini-columns.
                for k in np.arange(7): #this loops through each of the seven values
                    try:
                        output+=str(d[i][j][k])+'\t'
                    except IndexError:
                        output+='\t'
                output+='\t'
            output+='\n'
        f = open(filename, 'w')
        f.write(output)
        f.close()
        g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
    
measure = Measure()




# -*- coding: utf-8 -*-
import numpy as np
import scipy
import scipy.ndimage    
from skimage import feature, measure
from skimage.morphology import remove_small_objects
from skimage.filters import threshold_local
from qtpy import QtCore, QtGui, QtWidgets
from .. import global_vars as g
from .BaseProcess import BaseProcess, SliderLabel, WindowSelector,  MissingWindowError, CheckBox, ComboBox
from ..roi import makeROI, ROI_Drawing


__all__ = ['threshold','remove_small_blobs','adaptive_threshold','logically_combine','binary_dilation','binary_erosion', 'generate_rois', 'canny_edge_detector']
     
     
def convert2uint8(tif):
    oldmin = np.min(tif)
    oldmax = np.max(tif)
    newmax = 2**8-1
    tif = ((tif-oldmin)*newmax)/(oldmax-oldmin)
    tif = tif.astype(np.uint8)
    return tif
    
class Threshold(BaseProcess):
    """###threshold(value, darkBackground=False, keepSourceWindow=False)
    
Creates a boolean matrix by applying a threshold
    
    Parameters:
        | value (float) -- The threshold to be applied
        | darkBackground (bool) -- If this is True, pixels below the threshold will be True
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        valueSlider = SliderLabel(2)
        if g.currentWindow is not None:
            image=g.currentWindow.image
            valueSlider.setRange(np.min(image),np.max(image))
            valueSlider.setValue(np.mean(image))
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name': 'value','string': 'Value','object': valueSlider})
        self.items.append({'name': 'darkBackground', 'string':'Dark Background','object': CheckBox()})
        self.items.append({'name': 'preview','string': 'Preview','object': preview})
        super().gui()

    def __call__(self, value, darkBackground=False, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.oldwindow.nDims > 3:
            g.alert("You cannot run this function on an image of dimension greater than 3. If your window has color, convert to a grayscale image before running this function")
            return None
        if darkBackground:
            newtif = self.tif < value
        else:
            newtif = self.tif > value
        self.newtif = newtif.astype(np.uint8)
        self.newname = self.oldname+' - Thresholded '+str(value)
        return self.end()

    def preview(self):
        if g.currentWindow is None or g.currentWindow.closed:
            return
        win = g.currentWindow
        value = self.getValue('value')
        preview = self.getValue('preview')
        darkBackground = self.getValue('darkBackground')
        if win.nDims > 3:
            g.alert("You cannot run this function on an image of dimension greater than 3. If your window has color, convert to a grayscale image before running this function")
            return None
        if preview:
            if win.nDims == 3: # if the image is 3d
                testimage = np.copy(win.image[win.currentIndex])
            elif win.nDims == 2:
                testimage = np.copy(win.image)
            if darkBackground:
                testimage = testimage<value
            else:
                testimage = testimage>value
            win.imageview.setImage(testimage, autoLevels=False)
            win.imageview.setLevels(-.1,1.1)
        else:
            win.reset()
            if win.nDims == 3:
                image = win.image[win.currentIndex]
            else:
                image = win.image
            win.imageview.setLevels(np.min(image), np.max(image))
threshold = Threshold()

class BlocksizeSlider(SliderLabel):
    def __init__(self,demicals=0):
        SliderLabel.__init__(self,demicals)
    def updateSlider(self,value):
        if value%2==0:
            if value<self.slider.value():
                value-=1
            else:
                value+=1
            self.label.setValue(value)
        self.slider.setValue(int(value*10**self.decimals))
    def updateLabel(self,value):
        if value%2==0:
            value-=1
        self.label.setValue(value)
    
class Adaptive_threshold(BaseProcess):
    """###adaptive_threshold(value, block_size, darkBackground=False, keepSourceWindow=False)
    
Creates a boolean matrix by applying an adaptive threshold using the scikit-image threshold_local function
    
    Parameters:
        | value (int) -- The threshold to be applied
        | block_size (int)  -- size of a pixel neighborhood that is used to calculate a threshold value for the pixel. Must be an odd number greater than 3.
        | darkBackground (bool) -- If this is True, pixels below the threshold will be True
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        valueSlider=SliderLabel(2)
        valueSlider.setRange(-20,20)
        valueSlider.setValue(0)
        block_size=BlocksizeSlider(0)
        if g.currentWindow is not None:
            max_block = int(max([g.currentWindow.image.shape[-1],g.currentWindow.image.shape[-2]])/2)
        else:
            max_block = 100
        block_size.setRange(3, max_block)
        preview = CheckBox(); preview.setChecked(True)
        self.items.append({'name': 'value', 'string': 'Value', 'object': valueSlider})
        self.items.append({'name': 'block_size', 'string':'Block Size', 'object':block_size})
        self.items.append({'name': 'darkBackground', 'string': 'Dark Background', 'object': CheckBox()})
        self.items.append({'name': 'preview', 'string': 'Preview', 'object': preview})
        super().gui()
        self.preview()

    def __call__(self, value, block_size, darkBackground=False, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Local Threshold does not support float16 type arrays")
            return
        newtif = np.copy(self.tif)

        if self.oldwindow.nDims == 2:
            newtif = threshold_local(newtif, block_size, offset=value)
        elif self.oldwindow.nDims == 3:
            for i in np.arange(len(newtif)):
                newtif[i] = threshold_local(newtif[i], block_size, offset=value)
        else:
            g.alert("You cannot run this function on an image of dimension greater than 3. If your window has color, convert to a grayscale image before running this function")
            return None
        if darkBackground:
            newtif = np.logical_not(newtif)
        self.newtif = newtif.astype(np.uint8)
        self.newname = self.oldname + ' - Thresholded ' + str(value)
        return self.end()

    def preview(self):
        if g.currentWindow is None or g.currentWindow.closed:
            return
        win = g.currentWindow
        value = self.getValue('value')
        block_size = self.getValue('block_size')
        preview = self.getValue('preview')
        darkBackground = self.getValue('darkBackground')
        nDim = len(win.image.shape)
        if nDim > 3:
            g.alert("You cannot run this function on an image of dimension greater than 3. If your window has color, convert to a grayscale image before running this function")
            return None
        if preview:
            if nDim == 3: # if the image is 3d
                testimage=np.copy(win.image[win.currentIndex])
            elif nDim == 2:
                testimage=np.copy(win.image)
            testimage = threshold_local(testimage, block_size, offset=value)
            if darkBackground:
                testimage = np.logical_not(testimage)
            testimage = testimage.astype(np.uint8)
            win.imageview.setImage(testimage, autoLevels=False)
            win.imageview.setLevels(-.1, 1.1)
        else:
            win.reset()
            if nDim == 3:
                image = win.image[win.currentIndex]
            else:
                image = win.image
            win.imageview.setLevels(np.min(image), np.max(image))
adaptive_threshold=Adaptive_threshold()


class Canny_edge_detector(BaseProcess):
    """###canny_edge_detector(sigma, keepSourceWindow=False)
    
    Parameters:
        | sigma (float) -- 
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        sigma=SliderLabel(2)
        if g.currentWindow is not None:
            sigma.setRange(0,1000)
            sigma.setValue(1)
        preview=CheckBox(); preview.setChecked(True)
        self.items.append({'name':'sigma','string':'Sigma','object':sigma})
        self.items.append({'name':'preview','string':'Preview','object':preview})
        super().gui()
        self.preview()
    def __call__(self,sigma, keepSourceWindow=False):
        self.start(keepSourceWindow)
        nDim=len(self.tif.shape) 
        newtif=np.copy(self.tif)

        if self.tif.dtype == np.float16:
            g.alert("Canny Edge Detection does not work on float32 images. Change the data type to use this function.")
            return None

        if nDim==2:
            newtif=feature.canny(self.tif,sigma)
        else:
            for i in np.arange(len(newtif)):
                newtif[i] = feature.canny(self.tif[i],sigma)
        self.newtif=newtif.astype(np.uint8)
        self.newname=self.oldname+' - Canny '
        return self.end()

    def preview(self):
        if g.currentWindow is None or g.currentWindow.closed:
            return
        win = g.currentWindow
        sigma = self.getValue('sigma')
        preview = self.getValue('preview')
        nDim = len(win.image.shape)
        if preview:
            if nDim==3: # if the image is 3d
                testimage=np.copy(win.image[win.currentIndex])
            elif nDim==2:
                testimage=np.copy(win.image)
            testimage=feature.canny(testimage,sigma)
            win.imageview.setImage(testimage,autoLevels=False)
            win.imageview.setLevels(-.1,1.1)
        else:
            win.reset()
            if nDim==3:
                image=win.image[win.currentIndex]
            else:
                image=win.image
            win.imageview.setLevels(np.min(image),np.max(image))
canny_edge_detector=Canny_edge_detector()


class Logically_combine(BaseProcess):
    """###logically_combine(window1, window2,operator, keepSourceWindow=False)

Combines two windows according to the operator
    
    Parameters:
        | window1 (Window)
        | window2 (Window)
        | operator (str)
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        window1=WindowSelector()
        window2=WindowSelector()
        operator=ComboBox()
        operator.addItem('AND')
        operator.addItem('OR')
        operator.addItem('XOR')
        self.items.append({'name':'window1','string':'Window 1','object':window1})
        self.items.append({'name':'window2','string':'Window 2','object':window2})
        self.items.append({'name':'operator','string':'Operator','object':operator})
        super().gui()
    def __call__(self,window1, window2,operator,keepSourceWindow=False):
        self.keepSourceWindow=keepSourceWindow
        g.m.statusBar().showMessage('Performing {}...'.format(self.__name__))
        if window1 is None or window2 is None:
            raise(MissingWindowError("You cannot execute '{}' without selecting a window first.".format(self.__name__)))
        if window1.image.shape!=window2.image.shape:
            g.m.statusBar().showMessage('The two windows have images of different shapes. They could not be combined')
            return None
        if operator=='AND':
            self.newtif=np.logical_and(window1.image,window2.image)
        elif operator=='OR':
            self.newtif=np.logical_or(window1.image,window2.image)
        elif operator=='XOR':
            self.newtif=np.logical_xor(window1.image,window2.image)
            
        self.oldwindow=window1
        self.oldname=window1.name
        self.newname=self.oldname+' - Logical {}'.format(operator)
        if keepSourceWindow is False:
            window2.close()
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        return self.end()
logically_combine=Logically_combine()

    
class Remove_small_blobs(BaseProcess):
    """###remove_small_blobs(rank, value, keepSourceWindow=False)

Finds all contiguous 'True' pixels in rank dimensions.  Removes regions which have fewer than the specified pixels.
    
    Parameters:
        | rank  (int) -- The number of dimensions.  If rank==2, each frame is treated independently
        | value (int) -- The size (in pixels) below which each contiguous region must be in order to be discarded.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        rank = QtWidgets.QSpinBox()
        rank.setRange(2,3)
        value = QtWidgets.QSpinBox()
        value.setRange(1, 100000)
        self.items.append({'name': 'rank', 'string': 'Number of Dimensions', 'object': rank})
        self.items.append({'name': 'value', 'string': 'Value', 'object': value})
        super().gui()

    def __call__(self, rank, value, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("remove_small_blobs() does not support float16 type arrays")
            return
        oldshape = self.tif.shape
        newtif = np.zeros_like(self.tif, dtype='bool')
        if self.oldwindow.nDims == 2:
            newtif = remove_small_objects(self.tif.astype('bool'), value, connectivity=2)
        elif self.oldwindow.nDims == 3:
            if rank == 2:
                for i in np.arange(len(self.tif)):
                    newtif[i] = remove_small_objects(self.tif[i].astype('bool'), value, connectivity=2)
            elif rank == 3:
                newtif = remove_small_objects(self.tif.astype('bool'), value, connectivity=2)
        self.newtif = newtif
        self.newname = self.oldname + ' - Removed Blobs ' + str(value)
        return self.end()

    def get_init_settings_dict(self):
        s = dict()
        s['rank'] = 2
        s['value'] = 1
        return s

remove_small_blobs = Remove_small_blobs()


class Binary_Dilation(BaseProcess):
    """###binary_dilation(rank,connectivity,iterations, keepSourceWindow=False)

Performs a binary dilation on a binary image.  The 'False' pixels neighboring 'True' pixels become converted to 'True' pixels.
    
    Parameters:
        | rank (int) -- The number of dimensions to dilate. Can be either 2 or 3.  
        | connectivity (int) -- `connectivity` determines the distance to dilate.
             `connectivity` may range from 1 (no diagonal elements are neighbors) 
             to `rank` (all elements are neighbors).
        | iterations (int) -- How many times to repeat the dilation
        | keepSourceWindow (bool) -- If this is False, a new Window is created with the result. Otherwise, the currentWindow is used
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        rank=QtWidgets.QSpinBox()
        rank.setRange(2,3)
        connectivity=QtWidgets.QSpinBox()
        connectivity.setRange(1,3)
        iterations=QtWidgets.QSpinBox()
        iterations.setRange(1,100)
        self.items.append({'name':'rank','string':'Number of Dimensions','object':rank})
        self.items.append({'name':'connectivity','string':'Connectivity','object':connectivity})
        self.items.append({'name':'iterations','string':'Iterations','object':iterations})

        super().gui()
    def __call__(self,rank,connectivity,iterations, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Adaptive Threshold does not support float16 type arrays")
            return
        if len(self.tif.shape)==3 and rank==2:
            s=scipy.ndimage.generate_binary_structure(3,connectivity)
            s[0]=False
            s[2]=False
        else:
            s=scipy.ndimage.generate_binary_structure(rank,connectivity)
        self.newtif=scipy.ndimage.morphology.binary_dilation(self.tif,s,iterations)
        self.newtif=self.newtif.astype(np.uint8)
        self.newname=self.oldname+' - Dilated '
        return self.end()
binary_dilation=Binary_Dilation()


class Binary_Erosion(BaseProcess):
    """###binary_erosion(rank,connectivity,iterations, keepSourceWindow=False)

Performs a binary erosion on a binary image.  The 'True' pixels neighboring 'False' pixels become converted to 'False' pixels.
    
    Parameters:
        | rank (int) -- The number of dimensions to erode. Can be either 2 or 3.  
        | connectivity (int) -- `connectivity` determines the distance to erode. `connectivity` may range from 1 (no diagonal elements are neighbors) to `rank` (all elements are neighbors).
        | iterations (int) -- How many times to repeat the erosion
        | keepSourceWindow (bool) -- If this is False, a new Window is created with the result. Otherwise, the currentWindow is used
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        rank=QtWidgets.QSpinBox()
        rank.setRange(2,3)
        connectivity=QtWidgets.QSpinBox()
        connectivity.setRange(1,3)
        iterations=QtWidgets.QSpinBox()
        iterations.setRange(1,100)
        self.items.append({'name':'rank','string':'Number of Dimensions','object':rank})
        self.items.append({'name':'connectivity','string':'Connectivity','object':connectivity})
        self.items.append({'name':'iterations','string':'Iterations','object':iterations})

        super().gui()
    def __call__(self, rank, connectivity, iterations, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Binary Erosion does not work on float32 images. Change the data type to use this function.")
            return None
        if len(self.tif.shape)==3 and rank==2:
            s = scipy.ndimage.generate_binary_structure(3, connectivity)
            s[0] = False
            s[2] = False
        else:
            s = scipy.ndimage.generate_binary_structure(rank, connectivity)
        self.newtif = scipy.ndimage.morphology.binary_erosion(self.tif, s, iterations)
        self.newtif = self.newtif.astype(np.uint8)
        self.newname = self.oldname+' - Dilated '
        return self.end()
binary_erosion=Binary_Erosion()


class Generate_ROIs(BaseProcess):
    """###generate_rois(level, keepSourceWindow=False)

Uses a binary image to create ROIs from positive clusters.
    
    Parameters:
        | level (float) - value in [0, 1] to use when finding contours
        | keepSourceWindow (bool) -- If this is False, a new Window is created with the result. Otherwise, the currentWindow is used
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
        self.ROIs = []
    def gui(self):
        self.gui_reset()
        self.previewing = False
        self.toPreview = False
        level=SliderLabel(2)
        level.setRange(0,1)
        level.setValue(.5)
        minDensity=QtWidgets.QSpinBox()
        minDensity.setRange(4, 1000)
        self.items.append({'name':'level','string':'Contour Level','object':level})
        self.items.append({'name':'minDensity','string':'Minimum Density','object':minDensity})
        self.ROIs = []
        super().gui()
        self.ui.rejected.connect(self.removeROIs)

    def removeROIs(self):
        for roi in self.ROIs:
            roi.cancel()
        self.ROIs = []

    def __call__(self, level, minDensity, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Adaptive Threshold does not support float16 type arrays")
            return
        for roi in self.ROIs:
            roi.cancel()
        self.ROIs = []

        im = g.currentWindow.image if g.currentWindow.image.ndim == 2 else g.currentWindow.image[g.currentWindow.currentIndex]
        im = scipy.ndimage.morphology.binary_closing(im)
        if np.any(im < 0) or np.any(im > 1):
            raise Exception("The current image is not a binary image. Threshold first")

        thresholded_image = np.squeeze(im)
        labelled=measure.label(thresholded_image)
        ROIs = []
        for i in range(1, np.max(labelled)+1):
            if np.sum(labelled == i) >= minDensity:
                im = scipy.ndimage.morphology.binary_dilation(scipy.ndimage.morphology.binary_closing(labelled == i))
                outline_coords = measure.find_contours(im, level)
                if len(outline_coords) == 0:
                    continue
                outline_coords = outline_coords[0]
                new_roi = makeROI("freehand", outline_coords)
                ROIs.append(new_roi)

    def preview(self):
        if g.currentWindow is None or g.currentWindow.closed:
            return
        win = g.currentWindow
        if self.previewing:
            self.toPreview = True
            return
        self.previewing = True
        im = win.image if win.image.ndim == 2 else win.image[win.currentIndex]
        im = scipy.ndimage.morphology.binary_closing(im)
        if np.any(im < 0) or np.any(im > 1):
            raise Exception("The current image is not a binary image. Threshold first")

        level = self.getValue('level')
        minDensity = self.getValue('minDensity')
        thresholded_image = np.squeeze(im)
        labelled=measure.label(thresholded_image)
        for roi in self.ROIs:
            roi.cancel()
        self.ROIs = []

        for i in range(1, np.max(labelled)+1):
            QtWidgets.qApp.processEvents()
            if np.sum(labelled == i) >= minDensity:
                im = scipy.ndimage.morphology.binary_dilation(scipy.ndimage.morphology.binary_closing(labelled == i))
                outline_coords = measure.find_contours(im, level)
                if len(outline_coords) == 0:
                    continue
                outline_coords = outline_coords[0]
                self.ROIs.append(ROI_Drawing(win, outline_coords[0][0], outline_coords[0][1], 'freehand'))
                for p in outline_coords[1:]:
                    self.ROIs[-1].extend(p[0], p[1])
                    QtWidgets.qApp.processEvents()

        self.previewing = False
        if self.toPreview:
            self.toPreview = False
            self.preview()

generate_rois = Generate_ROIs()













    
    
    


import sys
from qtpy import QtCore, QtGui

def format(color, style=''):
    """Return a QtGui.QTextCharFormat with the given attributes.
    """
    _color = QtGui.QColor()
    _color.setNamedColor(color)

    _format = QtGui.QTextCharFormat()
    _format.setForeground(_color)
    if 'bold' in style:
        _format.setFontWeight(QtGui.QFont.Bold)
    if 'italic' in style:
        _format.setFontItalic(True)
    return _format


# Syntax styles that can be shared by all languages
STYLES = {
    'keyword': format('blue'),
    'operator': format('red'),
    'brace': format('black'),
    'defclass': format('black', 'bold'),
    'string': format('magenta'),
    'string2': format('darkMagenta'),
    'comment': format('darkGreen', 'italic'),
    'self': format('black', 'italic'),
    'numbers': format('brown'),
}


class PythonHighlighter (QtGui.QSyntaxHighlighter):
    """Syntax highlighter for the Python language.
    """
    # Python keywords
    keywords = [
        'and', 'assert', 'break', 'class', 'continue', 'def',
        'del', 'elif', 'else', 'except', 'exec', 'finally',
        'for', 'from', 'global', 'if', 'import', 'in',
        'is', 'lambda', 'not', 'or', 'pass', 'print',
        'raise', 'return', 'try', 'while', 'yield',
        'None', 'True', 'False',
    ]

    # Python operators
    operators = [
        '=',
        # Comparison
        '==', '!=', '<', '<=', '>', '>=',
        # Arithmetic
        '\+', '-', '\*', '/', '//', '\%', '\*\*',
        # In-place
        '\+=', '-=', '\*=', '/=', '\%=',
        # Bitwise
        '\^', '\|', '\&', '\~', '>>', '<<',
    ]

    # Python braces
    braces = [
        '\{', '\}', '\(', '\)', '\[', '\]',
    ]
    def __init__(self, document):
        QtGui.QSyntaxHighlighter.__init__(self, document)

        # Multi-line strings (expression, flag, style)
        # FIXME: The triple-quotes in these two lines will mess up the
        # syntax highlighting from this point onward
        self.tri_single = (QtCore.QRegExp("'''"), 1, STYLES['string2'])
        self.tri_double = (QtCore.QRegExp('"""'), 2, STYLES['string2'])

        rules = []

        # Keyword, operator, and brace rules
        rules += [(r'\b%s\b' % w, 0, STYLES['keyword'])
            for w in PythonHighlighter.keywords]
        rules += [(r'%s' % o, 0, STYLES['operator'])
            for o in PythonHighlighter.operators]
        rules += [(r'%s' % b, 0, STYLES['brace'])
            for b in PythonHighlighter.braces]

        # All other rules
        rules += [
            # 'self'
            (r'\bself\b', 0, STYLES['self']),

            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, STYLES['string']),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, STYLES['string']),

            # 'def' followed by an identifier
            (r'\bdef\b\s*(\w+)', 1, STYLES['defclass']),
            # 'class' followed by an identifier
            (r'\bclass\b\s*(\w+)', 1, STYLES['defclass']),

            # From '#' until a newline
            (r'#[^\n]*', 0, STYLES['comment']),

            # Numeric literals
            (r'\b[+-]?[0-9]+[lL]?\b', 0, STYLES['numbers']),
            (r'\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b', 0, STYLES['numbers']),
            (r'\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b', 0, STYLES['numbers']),
        ]

        # Build a QtCore.QRegExp for each pattern
        self.rules = [(QtCore.QRegExp(pat), index, fmt)
            for (pat, index, fmt) in rules]


    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text.
        """
        # Do other syntax formatting
        for expression, nth, format in self.rules:
            index = expression.indexIn(text, 0)

            while index >= 0:
                # We actually want the index of the nth match
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)


    def match_multiline(self, text, delimiter, in_state, style):
        """Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QtCore.QRegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """
        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False


# -*- coding: utf-8 -*-
import os
import pkg_resources

__all__ = ['image_path']

def image_path(image_name):
    """
    Return the absolute path to an image

    Parameters
    ----------
    image_name : str
       Name of image

    Returns
    -------
    path : str
      Full path to image
    """


    try:
        if pkg_resources.resource_exists('flika.images', image_name):
            return pkg_resources.resource_filename('flika.images', image_name)
        else:
            raise RuntimeError("image does not exist: %s" % image_name)
    except NotImplementedError:  # workaround for mac app
        result = os.path.dirname(__file__)
        return os.path.join(result.replace('site-packages.zip', 'flika'),
                            image_name)


# -*- coding: utf-8 -*-
import numpy as np
import skimage
import skimage.filters
from qtpy import QtWidgets, QtGui, QtCore
import time
from .. import global_vars as g
from .BaseProcess import BaseProcess, SliderLabel, SliderLabelOdd, CheckBox
from .progress_bar import ProgressBar

__all__ = ['gaussian_blur', 'difference_of_gaussians', 'mean_filter', 'variance_filter', 'median_filter','butterworth_filter','boxcar_differential_filter','wavelet_filter','difference_filter', 'fourier_filter', 'bilateral_filter']
###############################################################################
##################   SPATIAL FILTERS       ####################################
###############################################################################
class Gaussian_blur(BaseProcess):
    """ gaussian_blur(sigma, norm_edges=False, keepSourceWindow=False)

    This applies a spatial gaussian_blur to every frame of your stack.  
    
    Args:
        sigma (float): The width of the gaussian
        norm_edges (bool): If true, this reduces the values of the pixels near the edges so they have the same standard deviation as the rest of the image

    Returns:
        flika.window.Window

    """
    def __init__(self):
        super().__init__()
        assert 'gaussian' in skimage.filters.__dict__  # Make sure your version of skimage is >= 0.12.3
    def gui(self):
        self.gui_reset()
        sigma=SliderLabel(2)
        sigma.setRange(0,100)
        sigma.setValue(1)
        norm_edges=CheckBox()
        norm_edges.setChecked(False)
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name': 'sigma', 'string': 'Sigma (pixels)', 'object': sigma})
        self.items.append({'name': 'norm_edges', 'string': 'Normalize Edges', 'object': norm_edges})
        self.items.append({'name': 'preview', 'string': 'Preview', 'object': preview})
        super().gui()
        self.preview()

    def __call__(self, sigma, norm_edges=False, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if norm_edges == True:
            mode = 'constant'
        else:
            mode = 'nearest'
        if sigma>0:
            self.newtif=np.zeros(self.tif.shape)
            if len(self.tif.shape)==3:
                for i in np.arange(len(self.newtif)):
                    self.newtif[i]=skimage.filters.gaussian(self.tif[i].astype(np.float64),sigma, mode=mode)
            elif len(self.tif.shape)==2:
                self.newtif=skimage.filters.gaussian(self.tif.astype(np.float64),sigma, mode=mode)
            self.newtif=self.newtif.astype(g.settings['internal_data_type'])
        else:
            self.newtif=self.tif
        self.newname=self.oldname+' - Gaussian Blur sigma='+str(sigma)
        return self.end()
    def preview(self):
        norm_edges = self.getValue('norm_edges')
        if norm_edges == True:
            mode = 'constant'
        else:
            mode = 'nearest'
        sigma=self.getValue('sigma')
        preview=self.getValue('preview')
        if preview:
            if len(g.currentWindow.image.shape)==3:
                testimage=g.currentWindow.image[g.currentWindow.currentIndex].astype(np.float64)
            elif len(g.currentWindow.image.shape)==2:
                testimage=g.currentWindow.image.astype(np.float64)
            if sigma>0:
                testimage=skimage.filters.gaussian(testimage,sigma, mode=mode)
            g.currentWindow.imageview.setImage(testimage,autoLevels=False)            
        else:
            g.currentWindow.reset()
gaussian_blur=Gaussian_blur()


class Difference_of_Gaussians(BaseProcess):
    """ gaussian_blur(sigma1, sigma2, keepSourceWindow=False)

    This subtracts one gaussian blurred image from another to spatially bandpass filter.

    Args:
        sigma1 (float): The width of the first gaussian
        sigma2 (float): The width of the first gaussian

    Returns:
        flika.window.Window

    """

    def __init__(self):
        super().__init__()
        assert 'gaussian' in skimage.filters.__dict__  # Make sure your version of skimage is >= 0.12.3

    def gui(self):
        self.gui_reset()
        sigma1 = SliderLabel(2)
        sigma1.setRange(0, 100)
        sigma1.setValue(1)
        sigma2 = SliderLabel(2)
        sigma2.setRange(0, 100)
        sigma2.setValue(2)
        preview = CheckBox()
        preview.setChecked(True)
        self.items.append({'name': 'sigma1', 'string': 'Sigma 1 (pixels)', 'object': sigma1})
        self.items.append({'name': 'sigma2', 'string': 'Sigma 2 (pixels)', 'object': sigma2})
        self.items.append({'name': 'preview', 'string': 'Preview', 'object': preview})
        super().gui()
        self.preview()

    def __call__(self, sigma1, sigma2, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if sigma1 > 0 and sigma2 > 0:
            self.newtif = np.zeros(self.tif.shape)
            if len(self.tif.shape) == 3:
                for i in np.arange(len(self.newtif)):
                    self.newtif[i] = skimage.filters.gaussian(self.tif[i].astype(np.float64), sigma1, mode='nearest') -\
                                     skimage.filters.gaussian(self.tif[i].astype(np.float64), sigma2, mode='nearest')
            elif len(self.tif.shape) == 2:
                self.newtif = skimage.filters.gaussian(self.tif.astype(np.float64), sigma1, mode='nearest') - \
                              skimage.filters.gaussian(self.tif.astype(np.float64), sigma2, mode='nearest')
            self.newtif = self.newtif.astype(g.settings['internal_data_type'])
        else:
            self.newtif = self.tif
        self.newname = self.oldname + ' - Difference of Gaussians ({} {})'.format(sigma1, sigma2)
        return self.end()

    def preview(self):
        sigma1 = self.getValue('sigma1')
        sigma2 = self.getValue('sigma2')
        preview = self.getValue('preview')
        if preview:
            if len(g.currentWindow.image.shape) == 3:
                testimage = g.currentWindow.image[g.currentWindow.currentIndex].astype(np.float64)
            elif len(g.currentWindow.image.shape) == 2:
                testimage = g.currentWindow.image.astype(np.float64)
            if sigma1 > 0 and sigma2 > 0:
                testimage = skimage.filters.gaussian(testimage, sigma1, mode='nearest') - skimage.filters.gaussian(testimage, sigma2, mode='nearest')
            g.currentWindow.imageview.setImage(testimage, autoLevels=False)
        else:
            g.currentWindow.reset()


difference_of_gaussians = Difference_of_Gaussians()
###############################################################################
##################   TEMPORAL FILTERS       ###################################
###############################################################################
from scipy.signal import butter, filtfilt
    

class Butterworth_filter(BaseProcess):
    """ butterworth_filter(filter_order, low, high, keepSourceWindow=False)
    This filters a stack in time.
    
    Parameters:
        | filter_order (int) -- The order of the butterworth filter (higher->steeper).
        | low (float) -- The low frequency cutoff.  Must be between 0 and 1 and must be below high.
        | high (float) -- The high frequency cutoff.  Must be between 0 and 1 and must be above low.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        filter_order=QtWidgets.QSpinBox()
        filter_order.setRange(1,10)
        low=SliderLabel(5)
        low.setRange(0,1)
        low.setValue(0)
        high=SliderLabel(5)
        high.setRange(0,1)
        high.setValue(1)
        low.valueChanged.connect(lambda low: high.setMinimum(low))
        high.valueChanged.connect(lambda high: low.setMaximum(high))
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'filter_order','string':'Filter Order','object':filter_order})
        self.items.append({'name':'low','string':'Low Cutoff Frequency','object':low})
        self.items.append({'name':'high','string':'High Cutoff Frequency','object':high})
        self.items.append({'name':'preview','string':'Preview','object':preview})        
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        if self.roi is None or g.currentTrace is None:
            preview.setChecked(False)
            preview.setEnabled(False)
        
    def __call__(self,filter_order,low,high,keepSourceWindow=False):
        if low==0 and high==1:
            return
        self.start(keepSourceWindow)
        if self.tif.ndim != 3:
            g.alert("Butterworth filter only works on 3-dimensional movies.")
            return
        if g.settings['multiprocessing']:
            self.newtif=butterworth_filter_multi(filter_order,low,high,g.currentWindow.image)
        else:
            self.newtif=np.zeros(self.tif.shape,dtype=g.settings.d['internal_data_type'])
            mt,mx,my=self.tif.shape
            b,a,padlen=self.makeButterFilter(filter_order,low,high)
            for i in np.arange(mx):
                for j in np.arange(my):
                    self.newtif[:, i, j]=filtfilt(b,a, self.tif[:, i, j], padlen=padlen)
        self.newname=self.oldname+' - Butter Filtered'
        return self.end()
        
    def preview(self):
        if g.currentTrace is not None:
            filter_order=self.getValue('filter_order')
            low=self.getValue('low')
            high=self.getValue('high')
            preview=self.getValue('preview')
            if self.roi is not None:
                if preview:
                    if (low==0 and high==1) or (low==0 and high==0):
                        self.roi.onRegionChangeFinished() #redraw roi without filter
                    else:
                        b,a,padlen=self.makeButterFilter(filter_order,low,high)
                        trace=self.roi.getTrace()
                        trace=filtfilt(b,a, trace, padlen=padlen)
                        roi_index=g.currentTrace.get_roi_index(self.roi)
                        g.currentTrace.update_trace_full(roi_index,trace) #update_trace_partial may speed it up
                else:
                    self.roi.redraw_trace()
    def makeButterFilter(self,filter_order,low,high):
        padlen=0
        if high==1: 
            if low==0: #if there is no temporal filter at all,
                return None,None,None
            else: #if only high pass temporal filter
                [b,a]= butter(filter_order,low,btype='highpass')
                padlen=3
        else:
            if low==0:
                [b,a]= butter(filter_order,high,btype='lowpass')
            else:
                [b,a]=butter(filter_order,[low,high], btype='bandpass')
            padlen=6
        return b,a,padlen
        
butterworth_filter=Butterworth_filter()

        
def butterworth_filter_multi(filter_order,low,high,tif):
    nThreads= g.settings['nCores']
    mt,mx,my=tif.shape
    block_ends=np.linspace(0,mx,nThreads+1).astype(np.int)
    data=[tif[:, block_ends[i]:block_ends[i+1],:] for i in np.arange(nThreads)] #split up data along x axis. each thread will get one.
    args=(filter_order,low,high)
    progress = ProgressBar(butterworth_filter_multi_inner, data, args, nThreads, msg='Performing Butterworth Filter')
    if progress.results is None or any(r is None for r in progress.results):
        result=None
    else:
        result=np.concatenate(progress.results,axis=1).astype(g.settings['internal_data_type'])
    return result
    

def butterworth_filter_multi_inner(q_results, q_progress, q_status, child_conn, args):
    data=child_conn.recv()
    status=q_status.get(True) #this blocks the process from running until all processes are launched
    if status=='Stop':
        q_results.put(None)
    
    def makeButterFilter(filter_order,low,high):
        padlen=0
        if high==1: 
            if low==0: #if there is no temporal filter at all,
                return None,None,None
            else: #if only high pass temporal filter
                [b,a]= butter(filter_order,low,btype='highpass')
                padlen=3
        else:
            if low==0:
                [b,a]= butter(filter_order,high,btype='lowpass')
            else:
                [b,a]=butter(filter_order,[low,high], btype='bandpass')
            padlen=6
        return b,a,padlen
        
    filter_order,low,high = args
    b,a,padlen=makeButterFilter(filter_order,low,high)
    mt,mx,my=data.shape
    result=np.zeros(data.shape,g.settings['internal_data_type'])
    nPixels=mx*my
    pixel=0
    percent=0
    for x in np.arange(mx):
        for y in np.arange(my):
            if not q_status.empty():
                stop=q_status.get(False)
                q_results.put(None)
                return
            pixel+=1
            if percent<int(100*pixel/nPixels):
                percent=int(100*pixel/nPixels)
                q_progress.put(percent)
            result[:, x, y]=filtfilt(b,a, data[:, x, y], padlen=padlen)
    q_results.put(result)

from scipy.ndimage.filters import convolve
class Mean_filter(BaseProcess):
    """ mean_filter(nFrames, keepSourceWindow=False)
    This filters a stack in time.
    
    Parameters:
        | nFrames (int) -- Number of frames to average
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        nFrames=SliderLabel(0)
        nFrames.setRange(1,100)
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'nFrames','string':'nFrames','object':nFrames})
        self.items.append({'name':'preview','string':'Preview','object':preview})        
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
            
    def __call__(self,nFrames,keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Mean Filter does not support float16 type arrays")
            return
        if self.tif.ndim != 3:
            g.alert("Mean Filter only supports 3-dimensional movies.")
            return
        self.newtif=convolve(self.tif,weights=np.full((nFrames,1,1),1.0/nFrames))        
        self.newname=self.oldname+' - Mean Filtered'
        return self.end()
    def preview(self):
        nFrames=self.getValue('nFrames')
        preview=self.getValue('preview')
        if self.roi is not None:
            if preview:
                if nFrames==1:
                    self.roi.redraw_trace() #redraw roi without filter
                else:
                    trace=self.roi.getTrace()
                    trace=convolve(trace,weights=np.full((nFrames),1.0/nFrames))        
                    roi_index=g.currentTrace.get_roi_index(self.roi)
                    g.currentTrace.update_trace_full(roi_index,trace) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
mean_filter=Mean_filter()

def varfilt(trace, nFrames):
    result = np.zeros_like(trace)
    mt = len(trace)
    for i in np.arange(mt):
        i0 = int(i-nFrames/2)
        i1 = int(i+nFrames/2)
        if i0 < 0:
            i0 = 0
        if i1 > len(trace):
            i1 = len(trace)
        result[i] = np.var(trace[i0:i1])
    return result

class Variance_filter(BaseProcess):
    """ variance_filter(nFrames, keepSourceWindow=False)
    This filters a stack in time.

    Parameters:
        | nFrames (int) -- Number of frames to take teh variance of
    Returns:
        newWindow
    """

    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        nFrames = SliderLabel(0)
        nFrames.setRange(1, 100)
        preview = CheckBox()
        preview.setChecked(True)
        self.items.append({'name': 'nFrames', 'string': 'nFrames', 'object': nFrames})
        self.items.append({'name': 'preview', 'string': 'Preview', 'object': preview})
        super().gui()
        self.roi = g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)

    def __call__(self, nFrames, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Variance filter does not support float16 type arrays")
            return
        if self.tif.ndim != 3:
            g.alert("Variance filter only supports 3-dimensional movies.")
            return
        self.newtif = np.zeros(self.tif.shape)
        _, mx, my = self.tif.shape
        for i in np.arange(my):
            for j in np.arange(mx):
                self.newtif[:, i, j] = varfilt(self.tif[:, i, j], nFrames)
        self.newname = self.oldname + ' - Variance Filtered'
        return self.end()

    def preview(self):
        nFrames = self.getValue('nFrames')
        preview = self.getValue('preview')
        if self.roi is not None:
            if preview:
                if nFrames == 1:
                    self.roi.redraw_trace()  # redraw roi without filter
                else:
                    trace = self.roi.getTrace()
                    trace = varfilt(trace, nFrames)
                    roi_index = g.currentTrace.get_roi_index(self.roi)
                    g.currentTrace.update_trace_full(roi_index, trace)  # update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()


variance_filter = Variance_filter()

from scipy.signal import medfilt
class Median_filter(BaseProcess):
    """ median_filter(nFrames, keepSourceWindow=False)
    This filters a stack in time.
    
    Parameters:
        | nFrames (int) -- Number of frames to average.  This must be an odd number
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        nFrames=SliderLabelOdd()
        nFrames.setRange(1,100)
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'nFrames','string':'nFrames','object':nFrames})
        self.items.append({'name':'preview','string':'Preview','object':preview})        
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
            
    def __call__(self, nFrames, keepSourceWindow=False):
        if nFrames%2 == 0: #if value is even:
            g.alert('median_filter only takes odd numbers.  Operation cancelled')
            return None
        self.start(keepSourceWindow)
        if self.tif.ndim < 3:
            g.alert("Median filter requires at least 3 dimensions. %d < 3" % self.tif.ndim)
            return
        mx=self.tif.shape[2]
        my=self.tif.shape[1]
        self.newtif = np.zeros(self.tif.shape)
        for i in np.arange(my):
            for j in np.arange(mx):
                self.newtif[:, i, j]=medfilt(self.tif[:, i, j], kernel_size=nFrames)      
        self.newname=self.oldname+' - Median Filtered'
        return self.end()
    def preview(self):
        nFrames=self.getValue('nFrames')
        preview=self.getValue('preview')
        if self.roi is not None:
            if preview:
                print(nFrames)
                if nFrames==1:
                    self.roi.redraw_trace() #redraw roi without filter
                elif nFrames%2==0: #if value is even
                    return None
                else:
                    trace=self.roi.getTrace()
                    trace=medfilt(trace,kernel_size=nFrames)
                    roi_index=g.currentTrace.get_roi_index(self.roi)
                    g.currentTrace.update_trace_full(roi_index,trace) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
                
median_filter=Median_filter()




from scipy.fftpack import fft, ifft, fftfreq
class Fourier_filter(BaseProcess):
    """ fourier_filter(frame_rate, low, high, loglogPreview, keepSourceWindow=False)
    I'm going to eventually plot the trace in the frequency domain inside this box so you can see where the power is.

    Parameters:
        | frame_rate (int) -- Frame Rate in Hz
        | low (float) -- Low cutoff frequency for the fourier filter
        | high (float) -- High cutoff frequency for fourier filter
        | loglogPreview (boolean) -- whether or not to plot frequency spectrum on log log axes
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        frame_rate=QtWidgets.QDoubleSpinBox()
        frame_rate.setRange(.01,1000)
        frame_rate.setValue(200)
        low=SliderLabel(5)
        low.setRange(0,1)
        low.setValue(0)
        high=SliderLabel(5)
        high.setRange(0,1)
        high.setValue(1)
        frame_rate.valueChanged.connect(self.frame_rate_changed)
        low.valueChanged.connect(lambda low: high.setMinimum(low))
        high.valueChanged.connect(lambda high: low.setMaximum(high))
        preview=CheckBox()
        preview.setChecked(True)
        loglogPreview=CheckBox()
        self.items.append({'name':'frame_rate','string':'Frame Rate (Hz)','object':frame_rate})
        self.items.append({'name':'low','string':'Low Cutoff Frequency','object':low})
        self.items.append({'name':'high','string':'High Cutoff Frequency','object':high})
        self.items.append({'name':'loglogPreview','string':'Plot frequency spectrum on log log axes','object':loglogPreview})    
        self.items.append({'name':'preview','string':'Preview','object':preview})        
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
            loglogPreview.setEnabled(False) 
    def __call__(self, frame_rate, low, high, loglogPreview, keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.dtype == np.float16:
            g.alert("Fourier transform does not support float16 movies.")
            return
        if self.tif.ndim != 3 or self.tif.shape[2] == 3:
            g.alert('Fourier transform requires 3 dimensional movies.')
            return
        if low==0 and high==frame_rate/2.0:
            return
        mt,mx,my=self.tif.shape
        W = fftfreq(mt, d=1.0/frame_rate)
        filt=np.ones((mt))
        filt[(np.abs(W)<low)] = 0
        filt[(np.abs(W)>high)] = 0
        self.newtif=np.zeros(self.tif.shape)
        for i in np.arange(my):
            for j in np.arange(mx):
                f_signal = fft(self.tif[:, i, j])
                f_signal*=filt
                self.newtif[:, i, j]=np.real(ifft(f_signal))
        self.newname=self.oldname+' - Fourier Filtered'
        return self.end()
    def preview(self):
        frame_rate=self.getValue('frame_rate')
        low=self.getValue('low')
        high=self.getValue('high')
        loglogPreview=self.getValue('loglogPreview')
        preview=self.getValue('preview')
        if self.roi is not None:
            if preview:
                if (low==0 and high==frame_rate/2.0) or (low==0 and high==0):
                    self.roi.redraw_trace() #redraw roi without filter
                else:
                    trace=self.roi.getTrace()
                    W = fftfreq(len(trace), d=1.0/frame_rate)
                    f_signal = fft(trace)
                    f_signal[(np.abs(W)<low)] = 0
                    f_signal[(np.abs(W)>high)] = 0
                    cut_signal=np.real(ifft(f_signal))
                    roi_index=g.currentTrace.get_roi_index(self.roi)
                    g.currentTrace.update_trace_full(roi_index,cut_signal) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
                
    def frame_rate_changed(self):
        low=[item for item in self.items if item['name']=='low'][0]['object']
        high=[item for item in self.items if item['name']=='high'][0]['object']
        frame_rate=[item for item in self.items if item['name']=='frame_rate'][0]['object']
        f=frame_rate.value()
        low.setRange(0.0,f/2.0)
        high.setRange(0.0,f/2.0)
        low.setValue(0)
        high.setValue(f/2.0)
""" This is demo code for plotting in the frequency domain, something I hoepfully will get around to implementing

from numpy import sin, linspace, pi
from pylab import plot, show, title, xlabel, ylabel, subplot
from scipy import fft, arange

def plotSpectrum(y,Fs):
    n = len(y) # length of the signal
    k = arange(n)
    W = fftfreq(len(y), d=1/Fs)

    Y = fft(y) # fft computing and normalization
    f_signal[(np.abs(W)>10)] = 0
    plot(abs(f_signal)[0:N/2])
    cut_signal=np.real(ifft(f_signal))
    p=plot(trace)
    p.plot(cut_signal,pen=pg.mkPen('r'))
    
    plot(frq,abs(f_signal),'r') # plotting the spectrum
    xlabel('Freq (Hz)')
    ylabel('|Y(freq)|')"""
        
fourier_filter=Fourier_filter()



class Difference_filter(BaseProcess):
    """ difference_filter(keepSourceWindow=False)
    subtracts each frame from the preceeding frame
    
    
    Parameters:
        | None
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        if super().gui()==False:
            return False
    def __call__(self,keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif=np.zeros(self.tif.shape)
        for i in np.arange(1,len(self.newtif)):
            self.newtif[i]=self.tif[i]-self.tif[i-1]
        self.newname=self.oldname+' - Difference Filtered'
        return self.end()
difference_filter=Difference_filter()

    
class Boxcar_differential_filter(BaseProcess):
    """ boxcar_differential_filter(minNframes, maxNframes, keepSourceWindow=False)
    Applies a Boxcar differential filter by comparing each frameat index I to the frames in range [I+minNframes, I+maxNframes]

    Parameters:
        | minNframes (int) -- The starting point of your boxcar window.
        | maxNframes (int) -- The ending point of your boxcar window.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        minNframes=SliderLabel(0)
        minNframes.setRange(1,100)
        maxNframes=SliderLabel(0)
        maxNframes.setRange(2,101)
        minNframes.valueChanged.connect(lambda minn: maxNframes.setMinimum(minn+1))
        maxNframes.valueChanged.connect(lambda maxx: minNframes.setMaximum(maxx-1))
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'minNframes','string':'Minimum Number of Frames','object':minNframes})
        self.items.append({'name':'maxNframes','string':'Maximum Number of Frames','object':maxNframes})
        self.items.append({'name':'preview','string':'Preview','object':preview})  
        if super().gui()==False:
            return False
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
    def __call__(self,minNframes,maxNframes,keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif=np.zeros(self.tif.shape)
        for i in np.arange(maxNframes,len(self.newtif)):
            self.newtif[i]=self.tif[i]-np.min(self.tif[i-maxNframes:i-minNframes],0)
        self.newname=self.oldname+' - Boxcar Differential Filtered'
        return self.end()
    def preview(self):
        minNframes=self.getValue('minNframes')
        maxNframes=self.getValue('maxNframes')
        preview=self.getValue('preview')
        if self.roi is not None:
            if preview:
                self.tif=self.roi.window.image
                (mt,mx,my)=self.tif.shape
                cnt=np.array([np.array([np.array([p[1],p[0]])]) for p in self.roi.pts ])
                mask=np.zeros(self.tif[0,:,:].shape,np.uint8)
                cv2.drawContours(mask,[cnt],0,255,-1)
                mask=mask.reshape(mx*my).astype(np.bool)
                tif=self.tif.reshape((mt,mx*my))
                tif=tif[:,mask]
                newtrace=np.zeros(mt)
                for i in np.arange(maxNframes,mt):
                    newtrace[i]=np.mean(tif[i]-np.min(tif[i-maxNframes:i-minNframes],0))
                roi_index=g.currentTrace.get_roi_index(self.roi)
                g.currentTrace.update_trace_full(roi_index,newtrace) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
boxcar_differential_filter=Boxcar_differential_filter()
    
    
from scipy import signal
class Wavelet_filter(BaseProcess):
    ''' wavelet_filter(low, high, keepSourceWindow=False)
    ***Warning!! This function is extremely slow.***
    
    Parameters:
        | low (int) -- The starting point of your boxcar window.
        | high (int) -- The ending point of your boxcar window.
    Returns:
        newWindow
    '''
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        low=SliderLabel(0)
        low.setRange(1,50)
        high=SliderLabel(0)
        high.setRange(2,50)
        low.valueChanged.connect(lambda minn: high.setMinimum(minn+1))
        high.valueChanged.connect(lambda maxx: low.setMaximum(maxx-1))
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'low','string':'Low Frequency Threshold','object':low})
        self.items.append({'name':'high','string':'High Frequency Threshold','object':high})
        self.items.append({'name':'preview','string':'Preview','object':preview})  
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
    def __call__(self,low,high,keepSourceWindow=False):
        self.start(keepSourceWindow)
        if self.tif.ndim != 3:
            g.alert("Wavelet filter only works on 3 dimensional movies")
            return
        mx=self.tif.shape[2]
        my=self.tif.shape[1]
        self.newtif=np.zeros(self.tif.shape)
        wavelet = signal.ricker
        widths = np.arange(low, high)
        for i in np.arange(my):
            print(i)
            for j in np.arange(mx):
                cwtmatr = signal.cwt(self.tif[:, i, j], wavelet, widths)
                self.newtif[:, i, j]=np.mean(cwtmatr,0)
        self.newname=self.oldname+' - Wavelet Filtered'
        return self.end()
    def preview(self):
        low=self.getValue('low')
        high=self.getValue('high')
        preview=self.getValue('preview')
        if self.roi is not None:
            if preview:
                trace=self.roi.getTrace()
                wavelet = signal.ricker
                widths = np.arange(low, high)
                cwtmatr = signal.cwt(trace, wavelet, widths)
                newtrace=np.mean(cwtmatr,0)
                roi_index=g.currentTrace.get_roi_index(self.roi)
                g.currentTrace.update_trace_full(roi_index,newtrace) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
wavelet_filter=Wavelet_filter()
    
    
    
    
    
    
    
    
    
    
class Bilateral_filter(BaseProcess):
    ''' bilateral_filter( keepSourceWindow=False)
    
    Parameters:
        | soft (bool)  --     True for guassian, False for hard filter
        | beta (float) --     beta of kernel
        | width (float) --    width of kernel
        | stoptol (float) --  tolerance for convergence
        | maxiter (int)  --   maximum number of iterations
    Returns:
        newWindow
    '''
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        
        soft=CheckBox()
        soft.setChecked(True)
        beta=SliderLabel(2)
        beta.setRange(1,500)
        beta.setValue(200)
        width=SliderLabel(2)
        width.setRange(1,50)
        width.setValue(8)
        stoptol=SliderLabel(4)
        stoptol.setRange(0,.02)
        stoptol.setValue(.001)
        maxiter=SliderLabel(0)
        maxiter.setRange(1,100)
        maxiter.setValue(10)
        preview=CheckBox()
        preview.setChecked(True)
        self.items.append({'name':'soft','string':'soft','object':soft})  
        self.items.append({'name':'beta','string':'beta','object':beta})  
        self.items.append({'name':'width','string':'width','object':width})  
        self.items.append({'name':'stoptol','string':'stop tolerance','object':stoptol})  
        self.items.append({'name':'maxiter','string':'Maximum Iterations','object':maxiter})  
        self.items.append({'name':'preview','string':'Preview','object':preview})  
        super().gui()
        self.roi=g.currentWindow.currentROI
        if self.roi is not None:
            self.ui.rejected.connect(self.roi.redraw_trace)
            self.ui.accepted.connect(self.roi.redraw_trace)
        else:
            preview.setChecked(False)
            preview.setEnabled(False)
        
    def __call__(self,soft, beta, width, stoptol, maxiter, keepSourceWindow=False):
        
        self.start(keepSourceWindow)
        if self.tif.ndim != 3:
            g.alert("Bilateral filter requires 3-dimensional image.")
            return
        if g.settings['multiprocessing']:
            self.newtif=bilateral_filter_multi(soft,beta,width,stoptol,maxiter,g.currentWindow.image)
        else:
            self.newtif=np.zeros(self.tif.shape)
            mt,mx,my=self.tif.shape
            for i in np.arange(mx):
                for j in np.arange(my):
                    self.newtif[:, i, j]=bilateral_smooth(soft,beta,width,stoptol,maxiter,self.tif[:,i,j])
        self.newname=self.oldname+' - Bilateral Filtered'
        return self.end()

    def preview(self):
        soft=self.getValue('soft')
        beta=self.getValue('beta')
        width=self.getValue('width')
        stoptol=self.getValue('stoptol')
        maxiter=self.getValue('maxiter')
        preview=self.getValue('preview')
        
        if self.roi is not None:
            if preview:
                trace=self.roi.getTrace()
                trace=bilateral_smooth(soft,beta,width,stoptol,maxiter,trace)
                roi_index=g.currentTrace.get_roi_index(self.roi)
                g.currentTrace.update_trace_full(roi_index,trace) #update_trace_partial may speed it up
            else:
                self.roi.redraw_trace()
        

def bilateral_filter_multi(soft,beta,width,stoptol,maxiter,tif):
    nThreads= g.settings['nCores']
    mt,mx,my=tif.shape
    block_ends=np.linspace(0,mx,nThreads+1).astype(np.int)
    data=[tif[:, block_ends[i]:block_ends[i+1],:] for i in np.arange(nThreads)] #split up data along x axis. each thread will get one.
    args=(soft,beta,width,stoptol,maxiter)
    progress = ProgressBar(bilateral_filter_inner, data, args, nThreads, msg='Performing Bilateral Filter')
    if progress.results is None or any(r is None for r in progress.results):
        result=None
    else:
        result=np.concatenate(progress.results,axis=1)
    return result
    
    
def bilateral_filter_inner(q_results, q_progress, q_status, child_conn, args):
    data=child_conn.recv() # unfortunately this step takes a long time
    percent=0  # This is the variable we send back which displays our progress
    status=q_status.get(True) #this blocks the process from running until all processes are launched
    if status=='Stop':
        q_results.put(None) # if the user presses stop, return None
    
    
    # Here is the meat of the inner_func.
    soft,beta,width,stoptol,maxiter=args #unpack all the variables inside the args tuple
    result=np.zeros(data.shape)
    tt,xx,yy=data.shape
    for x in np.arange(xx):
        for y in np.arange(yy):
            result[:,x,y]=bilateral_smooth(soft,beta,width,stoptol,maxiter,data[:,x,y])
            if not q_status.empty(): #check if the stop button has been pressed
                stop=q_status.get(False)
                q_results.put(None)
                return
        if percent<int(100*x/xx):
            percent=int(100*x/xx)
            q_progress.put(percent)
                    
    # finally, when we've finished with our calculation, we send back the result
    q_results.put(result)
    
def bilateral_smooth(soft,beta,width,stoptol,maxiter,y):
    display=False       # 1 to report iteration values
    
    y=np.array(y[:])
    N=np.size(y,0)
    w=np.zeros((N,N))
    j=np.arange(0,N)
    
    #construct initial bilateral kernel
    for i in np.arange(0,N):
        w[i,np.arange(0,N)]=(abs(i-j) <= width)
    
    #initial guess from input signal
    xold=np.copy(y)
    
    #new matrix for storing distances
    d=np.zeros((N,N))
    
    #fig1 = plt.plot(y)
    
    if (display):
        if (soft):
            print('Soft kernel')
        else:
            print('Hard kernel')
        print('Kernel parameters beta= %d, W= %d' % (beta,width))
        print('Iter# Change')
    
    #start iteration
    iterate=1
    gap=np.inf
    
    while (iterate < maxiter):
    
        if (display):
            print('%d %f'% (iterate,gap))
    
        # calculate paiwise distances for all points
        for i in np.arange(0,N):
            d[:,i] = (0.5 * (xold - xold[i]) ** 2)
        
        #create kernel
        if (soft):
            W=np.multiply(np.exp(-beta*d),w)
    
        else:
            W=np.multiply((d <= beta ** 2),w)
        
        #apply kernel to get weighted mean shift   
        xnew1=np.sum(np.multiply(np.transpose(W),xold), axis=1)
        xnew2=np.sum(W, axis=1)
        xnew=np.divide(xnew1,xnew2)
       
        #plt.plot(xnew)   
        
        #check for convergence
        gap=np.sum(np.square(xold-xnew))
    
        if (gap < stoptol):
            if (display):
                print('Converged in %d iterations' % iterate)
            break
    
        xold=np.copy(xnew)
        iterate+=1
    return xold
    
bilateral_filter=Bilateral_filter()






#from scipy import signal
#data=g.currentTrace.rois[0]['roi'].getTrace()
#wavelet = signal.ricker
#widths = np.arange(1, 200)
#cwtmatr = signal.cwt(data, wavelet, widths)
#import pyqtgraph as pg
#i=pg.image(cwtmatr.T)
#i.view.setAspectLocked(lock=True, ratio=cwtmatr.shape[0]/cwtmatr.shape[1]*20)


# -*- coding: utf-8 -*-
from inspect import getmembers, ismodule
import numpy as np
import scipy
import pyqtgraph as pg
from .. import process
from .. import global_vars as g
from .. import window
from ..roi import open_rois


def getnamespace():
    """
    This function gets the namespace for the script interpreter.  It goes through all the modules in the 'process' package and loads all the objects defined in their '__all__' variables.
    """
    
    d = {}
    for name, mod in getmembers(process):
        if ismodule(mod):
            for func in mod.__all__:
                d[func] = mod.__dict__[func]
    d['g'] = g
    d['np'] = np
    d['scipy'] = scipy
    d['pg'] = pg
    d['plot'] = pg.plot
    d['Window'] = window.Window
    d['open_rois'] = open_rois
    return d


# -*- coding: utf-8 -*-
import numpy as np
import skimage
from qtpy import QtWidgets
from .. import global_vars as g
from .BaseProcess import BaseProcess, CheckBox


__all__ = ['set_value']

class Set_value(BaseProcess):
    """ set_value(value, firstFrame, lastFrame, restrictToROI=False, restrictToOutside=False, keepSourceWindow=False)
    This sets the value from firstFrame to lastFrame to value.
    
    Parameters:
        | value (int) -- The desired value
        | firstFrame (int) -- The first frame whos value you are setting
        | lastFrame (int) -- The last frame whos value you are altering
        | restrictToROI (bool) -- Whether or not only the current ROI will be effected.
    Returns:
        newWindow
    """
    def __init__(self):
        super().__init__()
    def gui(self):
        self.gui_reset()
        value=QtWidgets.QDoubleSpinBox()
        if g.currentWindow is not None:
            value.setRange(-2**64,2**64)
            value.setValue(0)
            firstFrame=QtWidgets.QSpinBox()
            firstFrame.setRange(0,len(g.currentWindow.image)-1)
            lastFrame=QtWidgets.QSpinBox()
            lastFrame.setRange(0,len(g.currentWindow.image)-1)
            lastFrame.setValue(len(g.currentWindow.image)-1)
        self.items.append({'name': 'value',             'string': 'Value',                   'object': value})
        self.items.append({'name': 'firstFrame',        'string': 'First Frame',             'object': firstFrame})
        self.items.append({'name': 'lastFrame',         'string': 'Last Frame',              'object': lastFrame})
        self.items.append({'name': 'restrictToROI',     'string': 'Restrict to current ROI', 'object': CheckBox()})
        self.items.append({'name': 'restrictToOutside', 'string': 'Restrict to everything outside current ROI','object': CheckBox()})
        super().gui()
    def __call__(self,value,firstFrame,lastFrame,restrictToROI=False, restrictToOutside=False, keepSourceWindow=False):
        self.start(keepSourceWindow)
        self.newtif=np.copy(self.tif)
        nDim = len(self.tif.shape)
        if nDim == 3:
            mt, mx, my=self.tif.shape
        elif nDim == 2:
            mx, my = self.tif.shape
        if restrictToROI:
            roi=g.currentWindow.currentROI
            roi.pts=roi.getPoints()
            x=np.array([p[0] for p in roi.pts])
            y=np.array([p[1] for p in roi.pts])
            xx,yy=skimage.draw.polygon(x,y)
            inside_bounds=(xx>=0) & (yy>=0) & (xx<mx) & (yy<my)
            xx=xx[inside_bounds]
            yy=yy[inside_bounds]
            #mask=np.zeros((mx,my),np.bool)
            #mask[xx,yy]=True
            if nDim==2:
                self.newtif[xx,yy]=value
            elif nDim==3:
                for i in np.arange(firstFrame,lastFrame+1):
                    self.newtif[i][xx,yy]=value
        elif restrictToOutside:
            roi=g.currentWindow.currentROI
            roi.pts = roi.getPoints()
            x=np.array([p[0] for p in roi.pts])
            y=np.array([p[1] for p in roi.pts])
            xx,yy=skimage.draw.polygon(x,y)
            inside_bounds=(xx>=0) & (yy>=0) & (xx<mx) & (yy<my)
            xx=xx[inside_bounds]
            yy=yy[inside_bounds]
            mask=np.ones((mx,my),np.bool)
            mask[xx,yy]=False
            if nDim==2:
                self.newtif[mask]=value
            elif nDim==3:
                for i in np.arange(firstFrame,lastFrame+1):
                    self.newtif[i][mask]=value
        
        else:
            if nDim==2:
                self.newtif[:] = value
            elif nDim==3:
                self.newtif[firstFrame:lastFrame+1] = value
        self.newname=self.oldname+' - value set to '+str(value)
        return self.end()
set_value=Set_value()




# -*- coding: utf-8 -*-
import sys, time
from qtpy import QtCore, QtGui, QtWidgets
from multiprocessing import Process, Queue, cpu_count, Pipe
import os
import numpy as np

__all__ = []

tic=time.time()

class ProcessLauncher(QtCore.QThread):
    status_update=QtCore.Signal(str)
    def __init__(self,parent):
        QtCore.QThread.__init__(self)
        self.parent=parent
        self.nCores=parent.nCores
    def __del__(self):
        self.wait()
    def run(self):
        p=self.parent
        nCores=self.nCores

        for i in range(nCores):
            self.status_update.emit('Creating process {}/{}'.format(i+1,nCores))
            q_result=Queue()
            q_progress=Queue()
            q_status=Queue()
            parent_conn, child_conn=Pipe()
            p.q_results.append(q_result)
            p.q_progress.append(q_progress)
            p.q_status.append(q_status)
            p.pipes.append(parent_conn)
            p.processes.append(Process(target=p.outerfunc, args=(q_result, q_progress, q_status, child_conn, p.args )))

        started=[False for i in range(nCores)]
        for i in range(nCores):
            if p.stopPressed:
                break
            self.status_update.emit('Initializing process {}/{}'.format(i+1,nCores))
            p.processes[i].start()
            started[i]=True
        for i in range(nCores):
            if started[i]:
                if not p.stopPressed:
                    self.status_update.emit('Sending data to process {}/{}'.format(i+1,nCores))
                    tic=time.time()
                    p.pipes[i].send(p.data[i])
                    #print('Time it took to send data: {}'.format(time.time()-tic))
        for i in range(nCores):
            if started[i]:
                if not p.stopPressed:
                    self.status_update.emit('Starting process {}/{}'.format(i+1,nCores))
                    p.q_status[i].put('Start')
                else:
                    p.q_status[i].put('Stop')
            else: #if the the process was never started
                p.process_finished[i]=True
        self.status_update.emit(p.msg)
        
        

        
    
class ProgressBar(QtWidgets.QWidget):
    finished_sig=QtCore.Signal()
    def __init__(self, outerfunc, data, args, nCores, msg='Performing Operations', parent=None ):
        super(ProgressBar, self).__init__(parent)
        self.outerfunc=outerfunc
        self.data=data
        self.args=args
        self.nCores=nCores
        self.msg=msg
        
        # GUI
        self.label=QtWidgets.QLabel(msg)
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.progress_bars=[]
        self.button = QtWidgets.QPushButton('Stop')
        self.button.clicked.connect(self.handleButton)
        main_layout = QtWidgets.QGridLayout()
        main_layout.addWidget(self.label,0,0)
        main_layout.addWidget(self.button, 0, 1)
        for i in range(nCores):
            bar=QtWidgets.QProgressBar()
            bar.setMinimum(1)
            bar.setMaximum(100)
            main_layout.addWidget(bar, 1+i, 0)
            self.progress_bars.append(bar)
        self.setLayout(main_layout)
        self.setWindowTitle(msg)
        self.stopPressed = False
        self.show()
        QtWidgets.qApp.processEvents()
        
        self.results=[None for i in range(nCores)]
        self.process_finished=[False for i in range(nCores)]
        self.q_results=[]
        self.q_progress=[]
        self.q_status=[]
        self.pipes=[]
        self.processes=[]
        self.processLauncher=ProcessLauncher(self)
        self.processLauncher.status_update.connect(self.status_updated)
        self.processLauncher.start()
        
        self.timer=QtCore.QTimer()
        self.timer.timeout.connect(self.check_if_finished)
        self.timer.start(50)
        
        self.loop = QtCore.QEventLoop()
        self.finished=False
        self.finished_sig.connect(self.loop.quit)
        self.finished_sig.connect(self.update_finished_status)
        self.loop.exec_() # This blocks until the "finished" signal is emitted
        
        while self.finished is False: #the exec_() loop doesn't wait for loop.quit when running in spyder for some reason.  This is the workaround
            time.sleep(.01)
            QtWidgets.qApp.processEvents()
        self.close()
        
    def check_if_finished(self):
        for i in range(len(self.q_progress)):
            if not self.q_progress[i].empty():
                while not self.q_progress[i].empty():
                    percent=self.q_progress[i].get()
                self.progress_bars[i].setValue(percent)
            if not self.q_results[i].empty():
                self.progress_bars[i].setValue(100)
                self.results[i]=self.q_results[i].get()
                self.process_finished[i]=True
                self.processes[i].join(1)
        QtWidgets.qApp.processEvents()
        if all(self.process_finished):
            if any(r is None for r in self.results):
                self.results=None
            self.timer.timeout.disconnect()
            self.finished_sig.emit()
        if self.stopPressed:
            #print('STOPPPPP')
            for i in range(self.nCores):
                self.q_status[i].put('Stop')

    def handleButton(self):
        self.stopPressed=True
    
    def status_updated(self,msg):
        self.label.setText(msg)
    def update_finished_status(self):
        self.finished=True

    def clear_memory(self):
        for child in self.children():
            child.deleteLater()
        for attr in dir(self):
            try:
                delattr(self, attr)
            except Exception:
                pass
        
    def clear_memory(self):
        for child in self.children():
            child.deleteLater()
        for attr in dir(self):
            try:
                delattr(self, attr)
            except Exception:
                pass
        
#    def closeEvent(self, event):
#        for child in self.findChildren(QtGui.QDialog):
#            if child is not widget:
#                child.deleteLater()
#                    
#        if self.closed:
#            print('This window was already closed')
#            event.accept()
#        else:
#            self.closeSignal.emit()
#            if hasattr(self,'image'):
#                del self.image
#            self.imageview.setImage(np.zeros((2,2))) #clear the memory
#            self.imageview.close()
#            del self.imageview
#            g.m.setWindowTitle("flika")
#            if g.currentWindow==self:
#                g.currentWindow=None
#            if self in g.windows:
#                g.windows.remove(self)
#            self.closed=True
#            event.accept() # let the window close
        
        
        
'''
When using the Progress Bar, you need to write two functions:
    1) An outer function that takes an object like a numpy array, breaks it into blocks, and creates the ProgressBar object.
    2) An inner function that receives the chunks, performs the processing, and returns the results.

If you are using a progress bar in a plugin, make sure to write it in its own python file, or the threads may crash. 
The first function should look something like this:  
'''
def outer_func():
    # get original data and arguments that the inner function will receive
    original_data=np.random.random((1000,200,200))
    args=(1,)
    
    #break data into blocks
    nCores = cpu_count()
    
    block_ends = np.linspace(0,len(original_data),nCores+1).astype(np.int)
    data_blocks=[original_data[block_ends[i]:block_ends[i+1]] for  i in np.arange(nCores)] # each thread will get one element of this list.
    
    # create the ProgressBar object
    progress = ProgressBar(inner_func, data_blocks, args, nCores, msg='Performing my cool inner function')
    
    # Once the ProgressBar object finishes running, the results of all the 
    # inner function's computations will be stored in progress.results in a 
    # list.  You will have to merge the list. If the user presses the Stop
    # button, the progress.results will be set to None
    if progress.results is None or any(r is None for r in progress.results):
        result=None
    else:
        result=np.concatenate(progress.results,axis=0)
    return result
'''
The inner function will actually do the heavy lifting.  
Each process will be running the inner function.  Each inner function must 
take the same 5 arguments.  
'''
def inner_func(q_results, q_progress, q_status, child_conn, args):
    data=child_conn.recv() # unfortunately this step takes a long time
    percent=0  # This is the variable we send back which displays our progress
    status=q_status.get(True) #this blocks the process from running until all processes are launched
    if status=='Stop':
        q_results.put(None) # if the user presses stop, return None
    
    
    # Here is the meat of the inner_func.
    val,=args #unpack all the variables inside the args tuple
    result=np.zeros(data.shape)
    ii,jj,kk=data.shape
    for i in np.arange(ii):
        for j in np.arange(jj):
            for k in np.arange(kk):
                result[i,j,k]=data[i,j,k]+val
            
        if not q_status.empty(): #check if the stop button has been pressed
            stop=q_status.get(False)
            q_results.put(None)
            return
        if percent<int(100*i/ii):
            percent=int(100*i/ii)
            q_progress.put(percent)
                    
    # finally, when we've finished with our calculation, we send back the result
    q_results.put(result)
    
    
    

        
if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    result=outer_func()
    print(result)





